# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from sushy.resources import base
from sushy.resources import common

LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):
    delete_job_queue = common.ActionField("#DellJobService.DeleteJobQueue")


class DellJobService(base.ResourceBase):

    _actions = ActionsField('Actions')
    identity = base.Field('Id', required=True)

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a DellJobService.

        :param connector: A Connector instance
        :param identity: The identity of the DellJobService resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        """
        super().__init__(
            connector, identity, redfish_version, registries)

    def delete_jobs(self, job_ids=['JID_CLEARALL']):
        """Delete the given jobs, or all jobs.

        :param job_ids: a list of job ids to delete. Clearing all the
            jobs may be accomplished using the keyword JID_CLEARALL
            as the job_id.
        """
        target_uri = self._actions.delete_job_queue.target_uri
        for job_id in job_ids:
            LOG.debug('Deleting the job %s', job_id)
            payload = {'JobID': job_id}
            self._conn.post(target_uri,
                            data=payload)
            LOG.info('Deleted the job %s', job_id)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import copy
import enum
from importlib import resources
import io
import json
import logging
import zipfile

from sushy import exceptions
from sushy.resources import constants
from sushy.resources import oem
from sushy import utils


LOG = logging.getLogger(__name__)


class Field:
    """Definition for fields fetched from JSON."""

    def __init__(self, path, required=False, default=None,
                 adapter=lambda x: x):
        """Create a field definition.

        :param path: JSON field to fetch the value from. Either a string,
            or a list of strings in case of a nested field.
        :param required: whether this field is required. Missing required,
            but not defaulted, fields result in MissingAttributeError.
        :param default: the default value to use when the field is missing.
        :param adapter: a function to call to transform and/or validate
            the received value. UnicodeError, ValueError or TypeError from
            this call are reraised as MalformedAttributeError.
        """
        if not callable(adapter):
            raise TypeError("Adapter must be callable")

        if not isinstance(path, list):
            path = [path]

        elif not path:
            raise ValueError('Path cannot be empty')

        self._path = path
        self._required = required
        self._default = default
        self._adapter = adapter

    def _get_item(self, dct, key_or_callable, **context):
        if not callable(key_or_callable):
            return dct[key_or_callable]

        for candidate_key in dct:
            if key_or_callable(
                    candidate_key, value=dct[candidate_key], **context):
                return dct[candidate_key]

        raise KeyError(key_or_callable)

    def _load(self, body, resource, nested_in=None):
        """Load this field from a JSON object.

        :param body: parsed JSON body.
        :param resource: ResourceBase instance for which the field is loaded.
        :param nested_in: parent resource path (for error reporting only),
            must be a list of strings or None.
        :raises: MissingAttributeError if a required field is missing
            and not defaulted.
        :raises: MalformedAttributeError on invalid field value or type.
        :returns: loaded and verified value
        """
        name = self._path[-1]
        for path_item in self._path[:-1]:
            body = body.get(path_item, {})

        try:
            item = self._get_item(body, name)

        except KeyError:
            if self._required:
                path = (nested_in or []) + self._path

                if self._default is None:
                    raise exceptions.MissingAttributeError(
                        attribute='/'.join(path),
                        resource=resource.path)

                logging.warning(
                    'Applying default "%s" on required, but missing '
                    'attribute "%s"', self._default, path)

            # Do not run the adapter on the default value
            return self._default

        # NOTE(etingof): this is just to account for schema violation
        if item is None:
            return

        try:
            return self._adapter(item)

        except (UnicodeError, ValueError, TypeError) as exc:
            path = (nested_in or []) + self._path
            raise exceptions.MalformedAttributeError(
                attribute='/'.join(path),
                resource=resource.path,
                error=exc)


def _collect_fields(resource):
    """Collect fields from the JSON.

    :param resource: ResourceBase or CompositeField instance.
    :returns: generator of tuples (key, field)
    """
    for attr in dir(resource.__class__):
        field = getattr(resource.__class__, attr)
        if isinstance(field, Field):
            yield (attr, field)


class CompositeField(collections.abc.Mapping, Field, metaclass=abc.ABCMeta):
    """Base class for fields consisting of several sub-fields."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subfields = dict(_collect_fields(self))

    def _load(self, body, resource, nested_in=None):
        """Load the composite field.

        :param body: parent JSON body.
        :param resource: parent resource.
        :param nested_in: parent resource name (for error reporting only).
        :returns: a new object with sub-fields attached to it.
        """
        nested_in = (nested_in or []) + self._path
        value = super()._load(body, resource)
        if value is None:
            return None

        # We need a new instance, as this method is called a singleton instance
        # that is attached to a class (not instance) of a resource or another
        # CompositeField. We don't want to end up modifying this instance.
        instance = copy.copy(self)
        for attr, field in self._subfields.items():
            # Hide the Field object behind the real value
            setattr(instance, attr, field._load(value, resource, nested_in))

        return instance

    # Satisfy the mapping interface, see
    # https://docs.python.org/3/library/collections.abc.html#collections.abc.Mapping

    def __getitem__(self, key):
        if key in self._subfields:
            return getattr(self, key)
        else:
            raise KeyError(key)

    def __len__(self):
        return len(self._subfields)

    def __iter__(self):
        return iter(self._subfields)


class ListField(Field):
    """Base class for fields consisting of a list of several sub-fields."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subfields = dict(_collect_fields(self))

    def _load(self, body, resource, nested_in=None):
        """Load the field list.

        :param body: parent JSON body.
        :param resource: parent resource.
        :param nested_in: parent resource name (for error reporting only).
        :returns: a new list object containing subfields.
        """
        nested_in = (nested_in or []) + self._path
        values = super()._load(body, resource)
        if values is None:
            return None

        # Initialize the list that will contain each field instance
        instances = []
        for value in values:
            instance = copy.copy(self)
            for attr, field in self._subfields.items():
                # Hide the Field object behind the real value
                setattr(instance, attr, field._load(value,
                                                    resource,
                                                    nested_in))
            instances.append(instance)

        return instances

    def __getitem__(self, key):
        return getattr(self, key)


class LinksField(CompositeField):
    """Reference to linked resources."""
    oem_vendors = Field('Oem', adapter=list)


class DictionaryField(Field):
    """Base class for fields consisting of dictionary of several sub-fields."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subfields = dict(_collect_fields(self))

    def _load(self, body, resource, nested_in=None):
        """Load the dictionary.

        :param body: parent JSON body.
        :param resource: parent resource.
        :param nested_in: parent resource name (for error reporting only).
        :returns: a new dictionary object containing subfields.
        """
        nested_in = (nested_in or []) + self._path
        values = super()._load(body, resource)
        if values is None:
            return None

        instances = {}
        for key, value in values.items():
            instance_value = copy.copy(self)
            for attr, field in self._subfields.items():
                # Hide the Field object behind the real value
                setattr(instance_value, attr, field._load(value,
                                                          resource,
                                                          nested_in))
            instances[key] = instance_value

        return instances

    def __getitem__(self, key):
        return getattr(self, key)


class MappedField(Field):
    """Field taking real value from a mapping."""

    def __init__(self, field, mapping, required=False, default=None):
        """Create a mapped field definition.

        :param field: JSON field to fetch the value from. This can be either
            a string or a list of string. In the latter case, the value will
            be fetched from a nested object.
        :param mapping: a mapping to take values from, a dictionary or
            an enumeration.
        :param required: whether this field is required. Missing required,
            but not defaulted, fields result in MissingAttributeError.
        :param default: the default value to use when the field is missing.
            This value is not matched against the mapping.
        """
        if isinstance(mapping, type) and issubclass(mapping, enum.Enum):
            def adapter(value):
                try:
                    return mapping(value)
                except ValueError:
                    return None

        elif isinstance(mapping, collections.abc.Mapping):
            adapter = mapping.get
        else:
            raise TypeError("The mapping argument must be a mapping or "
                            "an enumeration")

        super().__init__(
            field, required=required, default=default, adapter=adapter)


class MappedListField(Field):
    """Field taking a list of values with a mapping for the values

    Given JSON {'field':['xxx', 'yyy']}, a sushy resource definition and
    mapping {'xxx':'a', 'yyy':'b'}, the sushy object to come out will be like
    resource.field = ['a', 'b']
    """

    def __init__(self, field, mapping, required=False, default=None):
        """Create a mapped list field definition.

        :param field: JSON field to fetch the list of values from.
        :param mapping: a mapping for the list elements.
        :param required: whether this field is required. Missing required,
            but not defaulted, fields result in MissingAttributeError.
        :param default: the default value to use when the field is missing.
        """
        if isinstance(mapping, type) and issubclass(mapping, enum.Enum):
            def adapter(value):
                try:
                    return mapping(value)
                except ValueError:
                    return None

        elif isinstance(mapping, collections.abc.Mapping):
            adapter = mapping.get
        else:
            raise TypeError("The mapping argument must be a mapping or "
                            "an enumeration")

        self._mapping_adapter = adapter
        super().__init__(
            field, required=required, default=default,
            adapter=lambda x: x)

    def _load(self, body, resource, nested_in=None):
        """Load the mapped list.

        :param body: parent JSON body.
        :param resource: parent resource.
        :param nested_in: parent resource name (for error reporting only).
        :returns: a new list object containing the mapped values.
        """
        nested_in = (nested_in or []) + self._path
        values = super()._load(body, resource)

        if values is None:
            return

        instances = [self._mapping_adapter(value) for value in values
                     if self._mapping_adapter(value) is not None]

        return instances


class MessageListField(ListField):
    """List of messages with details of settings update status"""

    message_id = Field('MessageId')
    """The key for this message which can be used
    to look up the message in a message registry
    """

    message = Field('Message')
    """Human readable message, if provided"""

    severity = MappedField('Severity', constants.Severity)
    """Severity of the error"""

    resolution = Field('Resolution')
    """Used to provide suggestions on how to resolve
    the situation that caused the error
    """

    _related_properties = Field('RelatedProperties')
    """List of properties described by the message"""

    message_args = Field('MessageArgs')
    """List of message substitution arguments for the message
    referenced by `message_id` from the message registry
    """


class FieldData:
    """Contains data to be used when constructing Fields"""

    def __init__(self, status_code, headers, json_doc):
        """Initializes the FieldData instance"""
        self._status_code = status_code
        self._headers = headers
        self._json_doc = json_doc

    @property
    def status_code(self):
        """The status code"""
        return self._status_code

    @property
    def headers(self):
        """The headers"""
        return self._headers

    @property
    def json_doc(self):
        """The parsed JSON body"""
        return self._json_doc


class AbstractDataReader(metaclass=abc.ABCMeta):

    def set_connection(self, connector, path):
        """Sets mandatory connection parameters

        :param connector: A Connector instance
        :param path: path of the resource
        """
        self._conn = connector
        self._path = path

    @abc.abstractmethod
    def get_data(self):
        """Based on data source get data and parse to JSON"""


class JsonDataReader(AbstractDataReader):
    """Gets the data from HTTP response given by path"""

    def get_data(self):
        """Gets JSON file from URI directly"""
        data = self._conn.get(path=self._path)
        try:
            json_data = data.json() if data.content else {}
        except Exception as exc:
            LOG.error("Unable to parse JSON in response. %(exc)s. The server "
                      "returned:\n%(data)s",
                      {'exc': exc, 'data': data.content})
            raise
        return FieldData(data.status_code, data.headers, json_data)


class JsonPublicFileReader(AbstractDataReader):
    """Loads the data from the Internet"""

    def get_data(self):
        """Get JSON file from full URI"""
        data = self._conn.get(self._path)

        return FieldData(data.status_code, data.headers, data.json())


class JsonArchiveReader(AbstractDataReader):
    """Gets the data from JSON file in archive"""

    def __init__(self, archive_file):
        """Initializes the reader

        :param archive_file: file name of JSON file in archive
        """
        self._archive_file = archive_file

    def get_data(self):
        """Gets JSON file from archive. Currently supporting ZIP only"""

        data = self._conn.get(path=self._path)
        if data.headers.get('content-type') == 'application/zip':
            try:
                archive = zipfile.ZipFile(io.BytesIO(data.content))
                json_data = json.loads(archive.read(self._archive_file)
                                       .decode(encoding='utf-8'))
                return FieldData(data.status_code, data.headers, json_data)
            except (zipfile.BadZipfile, ValueError) as e:
                raise exceptions.ArchiveParsingError(
                    path=self._path, error=e)
        else:
            LOG.error('Support for %(type)s not implemented',
                      {'type': data.headers['content-type']})

            return FieldData(data.status_code, data.headers, None)


class JsonPackagedFileReader(AbstractDataReader):
    """Gets the data from packaged file given by path"""

    def __init__(self, resource_package_name):
        """Initializes the reader

        :param resource_package: Python package/module name
        """
        self._resource_package_name = resource_package_name

    def get_data(self):
        """Gets JSON file from packaged file denoted by path"""

        ref = resources.files(self._resource_package_name).joinpath(self._path)
        with ref.open(encoding='utf-8') as fp:
            json_data = json.load(fp)
            return FieldData(None, None, json_data)


def get_reader(connector, path, reader=None):
    """Create and configure the reader.

    :param connector: A Connector instance
    :param path: sub-URI path to the resource.
    :param reader: Reader to use to fetch JSON data.
    :returns: the reader
    """
    if reader is None:
        reader = JsonDataReader()
    reader.set_connection(connector, path)

    return reader


class ResourceBase(metaclass=abc.ABCMeta):

    redfish_version = None
    """The Redfish version"""

    _oem_vendors = Field('Oem', adapter=list)
    """The list of OEM extension names for this resource."""

    links = LinksField('Links')

    _log_resource_body = True
    """Whether to log the whole resource body in debug mode."""

    def __init__(self,
                 connector,
                 path='',
                 redfish_version=None,
                 registries=None,
                 reader=None,
                 json_doc=None,
                 root=None):
        """A class representing the base of any Redfish resource

        Invokes the ``refresh()`` method of resource for the first
        time from here (constructor).
        :param connector: A Connector instance
        :param path: sub-URI path to the resource.
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param reader: Reader to use to fetch JSON data.
        :param json_doc: parsed JSON document in form of Python types.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        self._conn = connector
        self._path = path
        self._json = None
        self.redfish_version = redfish_version
        self._registries = registries
        # Note(deray): Indicates if the resource holds stale data or not.
        # Starting off with True and eventually gets set to False when
        # attribute values are fetched.
        self._is_stale = True

        self._reader = get_reader(connector, path, reader)
        self._root = root
        self.refresh(json_doc=json_doc)

    def _get_value(self, val):
        """Iterate through the input to get values for all attributes

        :param val: Either a value or a resource
        :returns: Attribute value, which may be a dictionary
        """
        if isinstance(val, dict):
            subfields = {}
            for key, s_val in val.items():
                subfields[key] = self._get_value(s_val)
            return subfields

        elif isinstance(val, list):
            return [self._get_value(val[i]) for i in range(len(val))]

        elif (isinstance(val, DictionaryField)
              or isinstance(val, CompositeField)
              or isinstance(val, ListField)):
            subfields = {}
            for attr, field in val._subfields.items():
                subfields[attr] = self._get_value(val.__getitem__(attr))
            return subfields

        return val

    def _get_registry(self, identity, language='en', description='registry'):
        """Get a registry with the given identity.

        :param identity: The registry identity.
        :param language: RFC 5646 language code for Message Registries.
            Indicates language of registry to be used. Defaults to 'en'.
        :param description: Human-readable description to use in logging.
        :returns: the corresponding registry object or None.
        """
        registries = self._registries
        if not registries:
            LOG.info('No %s is available', description)
            return None

        for key, registry in registries.items():
            if (registry
                    and identity in (key, registry.identity)):
                # NOTE(iurygregory): some registries may have "en-US"
                # as their language, in this case we can check if the
                # registry language starts with the requested language.
                registry_language = registry.language.lower().split('-', 1)[0]
                if (language != registry.language
                        and language.lower() != registry_language):
                    LOG.debug('Found %(descr)s but its language %(reg_lang)s '
                              'does not match the requested %(lang)s',
                              {'descr': description,
                               'lang': language,
                               'reg_lang': registry.language})
                    continue

                return registry

        avail = ', '.join(f'{reg.identity} ({reg.language})'
                          for reg in registries.values())
        LOG.info('%(descr)s %(registry)s not available for language %(lang)s; '
                 'available are: %(avail)s',
                 {'descr': description,
                  'registry': self._attribute_registry,
                  'lang': language,
                  'avail': avail})
        return None

    def _parse_attributes(self, json_doc):
        """Parse the attributes of a resource.

        Parsed JSON fields are set to `self` as declared in the class.

        :param json_doc: parsed JSON document in form of Python types
        :returns: dictionary of attribute/values after parsing
        """
        settings = {}
        for attr, field in _collect_fields(self):
            # Hide the Field object behind the real value
            setattr(self, attr, field._load(json_doc, self))

            # Get the attribute/value pairs that have been parsed
            settings[attr] = self._get_value(getattr(self, attr))

        return settings

    def _get_etag(self):
        """Returns the ETag of the HTTP request if any was specified.

        :returns ETag or None
        """
        return self._get_headers().get('ETag')

    def _get_headers(self):
        """Returns the HTTP headers of the request for the resource.

        :returns: dict of HTTP headers
        """
        return self._reader.get_data()._headers

    def _allow_patch(self):
        """Returns if the resource supports the PATCH HTTP method.

        If the resource supports the PATCH HTTP method for updates,
        it will return it in the Allow HTTP header.
        :returns: Boolean flag if PATCH is supported or not
        """
        allow_header = self._get_headers().get('Allow', '')
        methods = set([h.strip().upper() for h in allow_header.split(',')])
        return "PATCH" in methods

    def refresh(self, force=True, json_doc=None):
        """Refresh the resource

        Freshly retrieves/fetches the resource attributes and invokes
        ``_parse_attributes()`` method on successful retrieval.
        It is recommended not to override this method in concrete ResourceBase
        classes. Resource classes can place their refresh specific operations
        in ``_do_refresh()`` method, if needed. This method represents the
        template method in the paradigm of Template design pattern.

        :param force: if set to False, will only refresh if the resource is
            marked as stale, otherwise neither it nor its subresources will
            be refreshed.
        :param json_doc: parsed JSON document in form of Python types.
        :raises: ResourceNotFoundError
        :raises: ConnectionError
        :raises: HTTPError
        """
        # Note(deray): Don't re-fetch / invalidate the sub-resources if the
        # resource is "_not_ stale" (i.e. fresh) OR _not_ forced.
        if not self._is_stale and not force:
            return

        data_source = ""
        if json_doc:
            self._json = json_doc
            data_source = "from expanded document"
        else:
            self._json = self._reader.get_data().json_doc

        attributes = self._parse_attributes(self._json)
        LOG.debug('Received representation of %(type)s %(path)s%(source)s: '
                  '%(json)s',
                  {'type': self.__class__.__name__,
                   'path': self._path,
                   'source': data_source,
                   'json': (attributes if self._log_resource_body
                            else '<stripped>')})
        self._do_refresh(force)

        # Mark it fresh
        self._is_stale = False

    def _do_refresh(self, force):
        """Primitive method to be overridden by refresh related activities.

        Derived classes are supposed to override this method with the
        resource specific refresh operations to be performed. This is a
        primitive method in the paradigm of Template design pattern.

        As for the base implementation of this method the approach taken is:
        On refresh, all sub-resources are marked as stale. That means
        invalidate (or undefine) the exposed attributes for nested resources
        for fresh evaluation in subsequent calls to those exposed attributes.
        In other words greedy-refresh is not done for them, unless forced by
        ``force`` argument.

        :param force: should force refresh the resource and its sub-resources,
            if set to True.
        :raises: ResourceNotFoundError
        :raises: ConnectionError
        :raises: HTTPError
        """
        utils.cache_clear(self, force_refresh=force)

    def invalidate(self, force_refresh=False):
        """Mark the resource as stale, prompting refresh() before getting used.

        If ``force_refresh`` is set to True, then it invokes ``refresh()``
        on the resource.

        :param force_refresh: will invoke refresh on the resource,
            if set to True.
        :raises: ResourceNotFoundError
        :raises: ConnectionError
        :raises: HTTPError
        """
        self._is_stale = True
        if force_refresh:
            self.refresh()

    @property
    def oem_vendors(self):
        return list(
            set((self._oem_vendors or []) + (self.links.oem_vendors or []))
        )

    @property
    def json(self):
        return self._json

    @property
    def path(self):
        return self._path

    def clone_resource(self, new_resource, path=''):
        """Instantiate given resource using existing BMC connection context"""
        return new_resource(
            self._conn, path or self.path,
            redfish_version=self.redfish_version,
            reader=self._reader,
            root=self.root)

    @property
    def resource_name(self):
        return utils.camelcase_to_underscore_joined(self.__class__.__name__)

    def get_oem_extension(self, vendor):
        """Get the OEM extension instance for this resource by OEM vendor

        :param vendor: the OEM vendor string which is the vendor-specific
            extensibility identifier. Examples are 'Contoso', 'Hpe'.
            Possible value can be got from ``oem_vendors`` attribute.
        :returns: the Redfish resource OEM extension instance.
        :raises: OEMExtensionNotFoundError
        """
        return oem.get_resource_extension_by_vendor(
            self.resource_name, vendor, self)

    @property
    def registries(self):
        return self._registries

    @property
    def root(self):
        return self._root


class ResourceLinksBase(ResourceBase, metaclass=abc.ABCMeta):

    def __init__(self, connector, path, redfish_version=None, registries=None,
                 root=None):
        """A class representing the base of any Redfish resource collection

        It gets inherited from ``ResourceBase`` and invokes the base class
        constructor.
        :param connector: A Connector instance
        :param path: sub-URI path to the resource collection.
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(connector, path, redfish_version, registries,
                         root=root)
        LOG.debug('Received %(count)d member(s) for %(type)s %(path)s',
                  {'count': len(self.members_identities),
                   'type': self.__class__.__name__, 'path': self._path})

    @property
    @abc.abstractmethod
    def members_identities(self):
        """A sequence with members identities"""

    @property
    @abc.abstractmethod
    def _resource_type(self):
        """The resource class that the collection contains.

        Override this property to specify the resource class that the
        collection contains.
        """

    def get_member(self, identity):
        """Given the identity return a ``_resource_type`` object

        :param identity: The identity of the ``_resource_type``
        :returns: The ``_resource_type`` object
        :raises: ResourceNotFoundError
        """
        return self._resource_type(
            self._conn, identity, redfish_version=self.redfish_version,
            registries=self.registries,
            root=self.root)

    @utils.cache_it
    def get_members(self):
        """Return a list of ``_resource_type`` objects present in collection

        :returns: A list of ``_resource_type`` objects
        """
        return [self.get_member(id_) for id_ in self.members_identities]


class ResourceCollectionBase(ResourceLinksBase):

    name = Field('Name')
    """The name of the collection"""

    members_identities = Field('Members', default=[],
                               adapter=utils.get_members_identities)
    """A tuple with the members identities"""


class MutableResourceCollectionBase(ResourceCollectionBase):

    def _create_member(self, values):
        """Create a new member of this collection.

        :param values: Fields to set on creation.
        :return: Created resource or None if it was not returned by the server.
        """
        response = self._conn.post(self._path, data=values)
        self.invalidate(force_refresh=True)

        location = response.headers.get('Location')
        if location is None:
            return None

        return self._resource_type(
            self._conn, location,
            redfish_version=self.redfish_version,
            registries=self.registries,
            root=self.root)

    def delete_member(self, identity):
        """Delete the given member of the collection."""
        self._conn.delete(identity)
        self.invalidate(force_refresh=True)



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum


class ExportTarget(enum.Enum):
    """Export system config action constants"""

    ALL = 'ALL'
    """Export entire system configuration"""

    BIOS = 'BIOS'
    """Export BIOS related configuration"""

    IDRAC = 'IDRAC'
    """Export iDRAC related configuration"""

    NIC = 'NIC'
    """Export NIC related configuration"""

    RAID = 'RAID'
    """Export RAID related configuration"""


# Backward compatibility
EXPORT_TARGET_ALL = ExportTarget.ALL
EXPORT_TARGET_BIOS = ExportTarget.BIOS
EXPORT_TARGET_IDRAC = ExportTarget.IDRAC
EXPORT_TARGET_NIC = ExportTarget.NIC
EXPORT_TARGET_RAID = ExportTarget.RAID


class ResetType(enum.Enum):
    """iDRAC Reset reset type constants"""

    GRACEFUL = 'Graceful'
    """Perform a graceful shutdown followed by a restart of the system"""

    FORCE = 'Force'
    """Perform an immediate (non-graceful) shutdown, followed by a restart"""


# Backward compatibility
RESET_IDRAC_GRACEFUL_RESTART = ResetType.GRACEFUL
RESET_IDRAC_FORCE_RESTART = ResetType.FORCE


class ShutdownType(enum.Enum):
    """ImportSystemConfiguration ShutdownType values"""

    GRACEFUL = 'Graceful'
    """Graceful shutdown for Import System Configuration

    Will wait for the host up to 5 minutes to shut down before timing out. The
    operating system can potentially deny or ignore the graceful shutdown
    request.
    """

    FORCED = 'Forced'
    """Forced shutdown for Import System Configuration

    The host server will be powered off immediately. Should be used when it is
    safe to power down the host.
    """

    NO_REBOOT = 'NoReboot'
    """No reboot for Import System Configuration

    No shutdown performed. Explicit reboot is necessary to apply changes.
    """


# Backward compatibility
IMPORT_SHUTDOWN_GRACEFUL = ShutdownType.GRACEFUL
IMPORT_SHUTDOWN_FORCED = ShutdownType.FORCED
IMPORT_SHUTDOWN_NO_REBOOT = ShutdownType.NO_REBOOT


class ExportUse(enum.Enum):
    """ExportUse in ExportSystemConfiguration"""

    DEFAULT = 'Default'
    """Default export type

    Leaves some attributes commented out and requires user to enable them
    before they can be applied during import.
    """

    CLONE = 'Clone'
    """Clone export type suitable for cloning a 'golden' configuration.

    Compared to Default export type, more attributes are enabled and
    storage settings adjusted to aid in cloning process.
    """

    REPLACE = 'Replace'
    """Replace export type suited for replacing complete configuration.

    Compared to Clone export type, most attributes are enabled and storage
    settings adjusted to aid in the replace process.
    """


# Backward compatibility
EXPORT_USE_DEFAULT = ExportUse.DEFAULT
EXPORT_USE_CLONE = ExportUse.CLONE
EXPORT_USE_REPLACE = ExportUse.REPLACE


class IncludeInExport(enum.Enum):
    """IncludeInExport in ExportSystemConfiguration"""

    DEFAULT = 'Default'
    """Default for what to include in export.

    Does not include read-only attributes, and depending on Export Use,
    passwords are marked as ****** (for Default) or are set to default password
    values (for Clone and Replace).
    """

    READ_ONLY = 'IncludeReadOnly'
    """Includes read-only attributes.

    In addition to values included by Default option, this also includes
    read-only attributes that cannot be changed via Import and are provided for
    informational purposes only.
    """

    PASSWORD_HASHES = 'IncludePasswordHashValues'  # noqa:S105
    """Include password hashes.

    When using Clone or Replace, include password hashes, instead of default
    password. Can be used to replicate passwords across systems.
    """

    READ_ONLY_PASSWORD_HASHES =\
        'IncludeReadOnly,IncludePasswordHashValues'  # noqa:S105
    """Includes both read-only attributes and password hashes.

    INCLUDE_EXPORT_READ_ONLY and INCLUDE_EXPORT_PASSWORD_HASHES combined
    """


# Backward compatibility
INCLUDE_EXPORT_DEFAULT = IncludeInExport.DEFAULT
INCLUDE_EXPORT_READ_ONLY = IncludeInExport.READ_ONLY
INCLUDE_EXPORT_PASSWORD_HASHES = IncludeInExport.PASSWORD_HASHES
INCLUDE_EXPORT_READ_ONLY_PASSWORD_HASHES =\
    IncludeInExport.READ_ONLY_PASSWORD_HASHES



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import subprocess
import time
from urllib.parse import urlparse

import sushy
from sushy.oem.dell import asynchronous
from sushy.oem.dell import constants
from sushy.oem.dell.resources import attributes
from sushy.oem.dell.resources.manager import constants as mgr_cons
from sushy.oem.dell.resources.manager import idrac_card_service
from sushy.oem.dell.resources.manager import job_collection
from sushy.oem.dell.resources.manager import job_service
from sushy.oem.dell.resources.manager import lifecycle_service
from sushy.oem.dell import utils
from sushy.resources import base
from sushy.resources import common
from sushy.resources.oem import base as oem_base
from sushy.taskmonitor import TaskMonitor
from sushy import utils as sushy_utils


LOG = logging.getLogger(__name__)

# System Configuration Tag Constant
_SYSTEM_CONFIG_TAG = "SystemConfiguration"

# Response Code Constant
_RESPONSE_OK_CODE = 200

# Structure {'Component FQDD': (tuple of beginning of Attribute keys)}
_DESTRUCTIVE_CONF_KEYS = {
    'iDRAC.Embedded.1':
        ('IPv4Static', 'IPv6Static', 'IPv4.1#Enable', 'IPv4.1#DHCPEnable',
         'IPv6.1#Enable', 'IPv6.1#AutoConfig')}


class SharedParameters(base.CompositeField):
    allowed_target_values = base.Field('Target@Redfish.AllowableValues')


class ExportActionField(common.ActionField):
    shared_parameters = SharedParameters('ShareParameters')
    allowed_export_use_values = base.Field(
        'ExportUse@Redfish.AllowableValues', adapter=list)
    allowed_include_in_export_values = base.Field(
        'IncludeInExport@Redfish.AllowableValues', adapter=list)


class ImportActionField(common.ActionField):
    allowed_shutdown_type_values = base.Field(
        'ShutdownType@Redfish.AllowableValues', adapter=list)


class DellManagerActionsField(base.CompositeField):
    import_system_configuration = ImportActionField(
        lambda key, **kwargs: key.endswith(
            '#OemManager.ImportSystemConfiguration'))

    export_system_configuration = ExportActionField(
        lambda key, **kwargs: key.endswith(
            '#OemManager.ExportSystemConfiguration'))


class DellManagerExtension(oem_base.OEMResourceBase):
    _actions = DellManagerActionsField('Actions')

    ACTION_DATA = {
        'ShareParameters': {
            'Target': ['ALL']
        },
        'ImportBuffer': None
    }

    # NOTE(etingof): iDRAC job would fail if this XML has
    # insignificant whitespaces

    IDRAC_CONFIG_CD = """\
<SystemConfiguration>\
<Component FQDD="%s">\
<Attribute Name="ServerBoot.1#BootOnce">\
%s\
</Attribute>\
<Attribute Name="ServerBoot.1#FirstBootDevice">\
VCD-DVD\
</Attribute>\
</Component>\
</SystemConfiguration>\
"""

    IDRAC_CONFIG_FLOPPY = """\
<SystemConfiguration>\
<Component FQDD="%s">\
<Attribute Name="ServerBoot.1#BootOnce">\
%s\
</Attribute>\
<Attribute Name="ServerBoot.1#FirstBootDevice">\
VFDD\
</Attribute>\
</Component>\
</SystemConfiguration>\
"""

    IDRAC_MEDIA_TYPES = {
        sushy.VIRTUAL_MEDIA_FLOPPY: IDRAC_CONFIG_FLOPPY,
        sushy.VIRTUAL_MEDIA_CD: IDRAC_CONFIG_CD
    }

    RETRY_COUNT = 35
    RETRY_DELAY = 15

    _IDRAC_IS_READY_RETRIES = 96
    _IDRAC_IS_READY_RETRY_DELAY_SEC = 10

    @property
    def import_system_configuration_uri(self):
        return self._actions.import_system_configuration.target_uri

    @property
    def export_system_configuration_uri(self):
        return self._actions.export_system_configuration.target_uri

    @property
    @sushy_utils.cache_it
    def idrac_card_service(self):
        """Reference to `DelliDRACCardService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DelliDRACCardService"],
            is_collection=False)

        return idrac_card_service.DelliDRACCardService(
            self._conn, path, self.redfish_version, self.registries)

    @property
    @sushy_utils.cache_it
    def lifecycle_service(self):
        """Property to reference `DellLCService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DellLCService"],
            is_collection=False)

        return lifecycle_service.DellLCService(
            self._conn, path, self.redfish_version, self.registries)

    @property
    @sushy_utils.cache_it
    def job_service(self):
        """Property to reference `DellJobService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DellJobService"],
            is_collection=False)

        return job_service.DellJobService(
            self._conn, path, self.redfish_version, self.registries)

    @property
    @sushy_utils.cache_it
    def job_collection(self):
        """Property to reference `DellJobService` instance of this manager.

        """
        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "Jobs"], is_collection=False)

        return job_collection.DellJobCollection(
            self._conn, path, self.redfish_version, self.registries)

    def set_virtual_boot_device(self, device, persistent=False,
                                manager=None, system=None):
        """Set boot device for a node.

        Dell iDRAC Redfish implementation does not support setting
        boot device to virtual media via standard Redfish means.
        However, this still can be done via an OEM extension.

        :param device: Boot device. Values are vendor-specific.
        :param persistent: Whether to set next-boot, or make the change
            permanent. Default: False.
        :param manager: Manager of OEM extension. Optional.
        :param system: System of OEM extension. Optional.
        :raises: InvalidParameterValue if Dell OEM extension can't
            be used.
        :raises: ExtensionError on failure to perform requested
            operation.
        """
        try:
            idrac_media = self.IDRAC_MEDIA_TYPES[device]

        except KeyError:
            raise sushy.exceptions.InvalidParameterValue(
                error=f'Unknown or unsupported device {device}')

        idrac_media = idrac_media % (
            manager.identity if manager else self._parent_resource.identity,
            'Disabled' if persistent else 'Enabled')

        action_data = dict(self.ACTION_DATA, ImportBuffer=idrac_media)

        # TODO(etingof): figure out if on-time or persistent boot can at
        # all be implemented via this OEM call

        attempts = self.RETRY_COUNT
        rebooted = False

        while True:
            try:
                response = asynchronous.http_call(
                    self._conn, 'post',
                    self.import_system_configuration_uri,
                    data=action_data,
                    sushy_task_poll_period=1)

                LOG.info("Set boot device to %(device)s via "
                         "Dell OEM magic spell (%(retries)d "
                         "retries)", {'device': device,
                                      'retries': self.RETRY_COUNT - attempts})

                return response

            except sushy.exceptions.HTTPError as exc:

                LOG.warning(
                    'Dell OEM set boot device failed (attempts left '
                    '%d): %s', attempts, exc)

                errors = exc.body and exc.body.get(
                    '@Message.ExtendedInfo') or []

                found = False

                for error in errors:
                    message_id = error.get('MessageId')

                    LOG.warning('iDRAC error: %s',
                                error.get('Message', 'Unknown error'))

                    if constants.IDRAC_CONFIG_PENDING in message_id:
                        found = True
                        if not rebooted:
                            LOG.warning(
                                'Let\'s try to turn it off and on again... '
                                'This may consume one-time boot settings if '
                                'set previously!')
                            utils.reboot_system(system)
                            rebooted = True
                            break

                    elif constants.IDRAC_JOB_RUNNING in message_id:
                        found = True
                        pass

                else:
                    if found:
                        time.sleep(self.RETRY_DELAY)
                    else:
                        raise

                if not attempts:
                    LOG.error('Too many (%d) retries, bailing '
                              'out.', self.RETRY_COUNT)
                    raise

                attempts -= 1

    def get_allowed_export_target_values(self):
        """Get the allowed targets of export system configuration.

        :returns: A set of allowed values.
        """
        export_action = self._actions.export_system_configuration
        allowed_values = export_action.shared_parameters.allowed_target_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'target of export system configuration at %s',
                        self.path)
            return set(mgr_cons.ExportTarget)

        return {v for v in mgr_cons.ExportTarget if v.value in allowed_values}

    def get_allowed_export_use_values(self):
        """Get allowed export use values of export system configuration.

        :returns: A set of allowed export use values.
        """
        export_action = self._actions.export_system_configuration
        allowed_values = export_action.allowed_export_use_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'export use of export system configuration at %s',
                        self.path)
            return set(mgr_cons.ExportUse)

        return {v for v in mgr_cons.ExportUse if v.value in allowed_values}

    def get_allowed_include_in_export_values(self):
        """Get allowed include in export values of export system configuration.

        :returns: A set of allowed include in export values.
        """
        export_action = self._actions.export_system_configuration
        allowed_values = export_action.allowed_include_in_export_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'include in export of export system configuration at '
                        '%s', self.path)
            return set(mgr_cons.IncludeInExport)

        return {v for v in mgr_cons.IncludeInExport
                if v.value in allowed_values}

    def _export_system_configuration(
        self, target, export_use=mgr_cons.ExportUse.DEFAULT,
        include_in_export=mgr_cons.IncludeInExport.DEFAULT):
        """Export system configuration.

        It exports system configuration for specified target like NIC, BIOS,
        RAID and allows to configure purpose for export and what to include.

        :param target: Component of the system to export the
            configuration from. Can be the entire system.
            Valid values can be gotten from
            `get_allowed_export_system_config_values`.
        :param export_use: Export use. Optional, defaults to "Default".
            Valid values can be gotten from `get_allowed_export_use_values`.
        :param include_in_export: What to include in export. Optional. Defaults
            to "Default". Valid values can be gotten from
            `get_allowed_include_in_export_values`.
        :returns: Response object containing configuration details.
        :raises: InvalidParameterValueError on invalid target.
        :raises: ExtensionError on failure to perform requested
            operation
        """
        valid_allowed_targets = self.get_allowed_export_target_values()
        if target not in valid_allowed_targets:
            raise sushy.exceptions.InvalidParameterValueError(
                parameter='target', value=target,
                valid_values=valid_allowed_targets)

        allowed_export_use = self.get_allowed_export_use_values()
        if export_use not in allowed_export_use:
            raise sushy.exceptions.InvalidParameterValueError(
                parameter='export_use', value=export_use,
                valid_values=allowed_export_use)

        allowed_include_in_export = self.get_allowed_include_in_export_values()
        if include_in_export not in allowed_include_in_export:
            # Check if value contains comma and validate each item separately
            # Older iDRACs used to include comma separated option in
            # AllowableValues but got removed in newer versions violating
            # AllowableValues validation logic.
            all_items_valid = True
            if not isinstance(include_in_export, mgr_cons.IncludeInExport):
                all_items_valid = False
            else:
                items = include_in_export.value.split(',')
                for item in items:
                    if (mgr_cons.IncludeInExport(item)
                            not in allowed_include_in_export):
                        all_items_valid = False
                        break

            if not all_items_valid:
                raise sushy.exceptions.InvalidParameterValueError(
                    parameter='include_in_export', value=include_in_export,
                    valid_values=allowed_include_in_export)

        target = mgr_cons.ExportTarget(target).value
        export_use = mgr_cons.ExportUse(export_use).value
        include_in_export = mgr_cons.IncludeInExport(include_in_export).value

        action_data = {
            'ShareParameters': {
                'Target': [target]
            },
            'ExportFormat': "JSON",
            'ExportUse': export_use,
            'IncludeInExport': include_in_export
        }

        try:
            response = asynchronous.http_call(
                self._conn,
                'post',
                self.export_system_configuration_uri,
                data=action_data)

            LOG.info("Successfully exported system configuration "
                     "for %(target)s", {'target': target})

            return response

        except (sushy.exceptions.ExtensionError,
                sushy.exceptions.InvalidParameterValueError) as exc:
            LOG.error('Dell OEM export system configuration failed : %s', exc)
            raise

    def export_system_configuration(self, include_destructive_fields=True):
        """Export system configuration.

        Exports ALL targets for cloning and includes password hashes and
        read-only attributes.

        :param include_destructive_fields: Whether includes settings such as
            iDRAC static IP address that could lead to losing access to iDRAC
            if importing this configuration into another system. Default to
            True for backward compatibility. False recommended if unsure.
        :returns: Response object containing configuration details.
        :raises: InvalidParameterValueError on invalid target.
        :raises: ExtensionError on failure to perform requested
            operation
        """

        include_in_export = mgr_cons.IncludeInExport.READ_ONLY_PASSWORD_HASHES
        response = self._export_system_configuration(
            mgr_cons.ExportTarget.ALL,
            export_use=mgr_cons.ExportUse.CLONE,
            include_in_export=include_in_export)

        if (response.status_code == _RESPONSE_OK_CODE
                and not include_destructive_fields):
            conf = response.json()
            if _SYSTEM_CONFIG_TAG in conf.keys():
                for fqdd, values in _DESTRUCTIVE_CONF_KEYS.items():
                    for comp in conf[_SYSTEM_CONFIG_TAG]['Components']:
                        if comp['FQDD'] == fqdd:
                            attributes_copy = comp['Attributes'].copy()
                            for child in comp['Attributes']:
                                if child.get('Name').startswith(values):
                                    attributes_copy.remove(child)
                            comp['Attributes'] = attributes_copy
            response._content = json.dumps(conf).encode()

        return response

    def get_pxe_port_macs_bios(self, ethernet_interfaces_mac):
        """Get a list of pxe port MAC addresses for BIOS.

        :param ethernet_interfaces_mac: Dictionary of ethernet interfaces.
        :returns: List of pxe port MAC addresses.
        :raises: ExtensionError on failure to perform requested operation.
        """
        pxe_port_macs = []
        # Get NIC configuration
        nic_settings = self._export_system_configuration(
            target=mgr_cons.ExportTarget.NIC)

        if nic_settings.status_code != _RESPONSE_OK_CODE:
            error = (('An error occurred when attempting to export '
                     'the system configuration. Status code: %(code), '
                      'Error details: %(err)'),
                     {'code': nic_settings.status_code,
                      'err': nic_settings.__dict__})
            LOG.error(error)
            raise sushy.exceptions.ExtensionError(error=error)
        # Parse the exported system configuration for the NIC
        # ports that are set to PXE boot
        json_data = nic_settings.json()
        if _SYSTEM_CONFIG_TAG in json_data.keys():
            for root in json_data[_SYSTEM_CONFIG_TAG]['Components']:
                nic_id = root['FQDD']
                for child in root['Attributes']:
                    if child.get('Name') == "LegacyBootProto":
                        if child['Value'] == "PXE":
                            mac_address = ethernet_interfaces_mac[nic_id]
                            pxe_port_macs.append(mac_address)
            return pxe_port_macs

        else:
            error = ('Failed to get system configuration from response')
            LOG.error(error)
            raise sushy.exceptions.ExtensionError(error=error)

    def get_allowed_import_shutdown_type_values(self):
        """Get the allowed shutdown types of import system configuration.

        :returns: A set of allowed shutdown type values.
        """
        import_action = self._actions.import_system_configuration
        allowed_values = import_action.allowed_shutdown_type_values

        if not allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'shutdown type of import system configuration at %s',
                        self.path)
            return set(mgr_cons.ShutdownType)

        return {v for v in mgr_cons.ShutdownType
                if v.value in allowed_values}

    def import_system_configuration(self, import_buffer):
        """Imports system configuration.

        Caller needs to handle system reboot separately.

        :param import_buffer: Configuration data to be imported.
        :returns: Task monitor instance to watch for task completion
        """
        action_data = dict(self.ACTION_DATA, ImportBuffer=import_buffer)
        # Caller needs to handle system reboot separately to preserve
        # one-time boot settings.
        shutdown_type = mgr_cons.ShutdownType.NO_REBOOT

        allowed_shutdown_types = self.get_allowed_import_shutdown_type_values()
        if shutdown_type not in allowed_shutdown_types:
            raise sushy.exceptions.InvalidParameterValueError(
                parameter='shutdown_type', value=shutdown_type,
                valid_values=allowed_shutdown_types)

        action_data['ShutdownType'] = shutdown_type.value

        response = self._conn.post(self.import_system_configuration_uri,
                                   data=action_data)

        return TaskMonitor.from_response(
            self._conn, response, self.import_system_configuration_uri)

    def reset_idrac(self, wait=True, ready_wait_time=60):
        """Reset the iDRAC and wait for it to become ready.

        :param wait: Whether to return immediately or wait for iDRAC to
            become operational.
        :param ready_wait_time: Amount of time in seconds to wait before
            starting to check on the iDRAC's status.
        """
        self.idrac_card_service.reset_idrac()
        if not wait:
            return
        host = urlparse(self._conn._url).netloc
        LOG.debug("iDRAC %(host)s was reset, "
                  "waiting for return to operational state", {'host': host})
        self._wait_for_idrac(host, ready_wait_time)
        self._wait_until_idrac_is_ready(host, self._IDRAC_IS_READY_RETRIES,
                                        self._IDRAC_IS_READY_RETRY_DELAY_SEC)

    @property
    def attributes(self):
        paths = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DellAttributes"],
            is_collection=True)

        for path in paths:
            yield attributes.DellAttributes(self._conn, path,
                                            self.redfish_version,
                                            self.registries)

    def _wait_for_idrac_state(self, host, alive=True, ping_count=3,
                              retries=24):
        """Wait for iDRAC to become pingable or not pingable.

        :param host: Hostname or IP of the iDRAC interface.
        :param alive: True for pingable state and False for not pingable
            state.
        :param ping_count: Number of consecutive ping results, per
            'alive', for success.
        :param retries: Number of ping retries.
        :returns: True on reaching specified host ping state; otherwise,
            False.
        """
        if alive:
            ping_type = "pingable"
        else:
            ping_type = "not pingable"
        LOG.debug("Waiting for iDRAC %(host)s to become %(ping_type)s",
                  {'host': host, 'ping_type': ping_type})
        response_count = 0
        while retries > 0:
            response = self._ping_host(host)
            retries -= 1
            if response == alive:
                response_count += 1
                LOG.debug("iDRAC %(host)s is %(ping_type)s, "
                          "count=%(response_count)s",
                          {'host': host, 'ping_type': ping_type,
                           'response_count': response_count})
                if response_count == ping_count:
                    LOG.debug("Reached specified %(alive)s count for iDRAC "
                              "%(host)s", {'alive': alive, 'host': host})
                    return True
            else:
                response_count = 0
                if alive:
                    LOG.debug("iDRAC %(host)s is still not pingable",
                              {'host': host})
                else:
                    LOG.debug("iDRAC %(host)s is still pingable",
                              {'host': host})
            time.sleep(10)
        return False

    def _wait_for_idrac(self, host, post_pingable_wait_time):
        """Wait for iDRAC to transition from unpingable to pingable.

        :param host: Hostname or IP of the iDRAC interface.
        :param post_pingable_wait_time: Amount of time in seconds to
            wait after the host becomes pingable.
        :raises: ExtensionError on failure to perform requested
            operation.
        """
        state_reached = self._wait_for_idrac_state(
            host, alive=False, ping_count=2, retries=24)
        if not state_reached:
            error_msg = ("Timed out waiting iDRAC %(host)s to become not "
                         "pingable", {'host': host})
            LOG.error(error_msg)
            raise sushy.exceptions.ExtensionError(error=error_msg)
        LOG.debug("iDRAC %(host)s has become not pingable", {'host': host})
        state_reached = self._wait_for_idrac_state(host, alive=True,
                                                   ping_count=3, retries=24)
        if not state_reached:
            error_msg = ("Timed out waiting iDRAC %(host)s to become pingable",
                         {'host': host})
            LOG.error(error_msg)
            raise sushy.exceptions.ExtensionError(error=error_msg)
        LOG.debug("iDRAC %(host)s has become pingable", {'host': host})
        time.sleep(post_pingable_wait_time)

    def _wait_until_idrac_is_ready(self, host, retries, retry_delay):
        """Wait until the iDRAC is in a ready state.

        :param host: Hostname or IP of the iDRAC interface.
        :param retries: The number of times to check if the iDRAC is
            ready.
        :param retry_delay: The number of seconds to wait between
            retries.
        :raises: ExtensionError on failure to perform requested
            operation.
        """

        while retries > 0:
            LOG.debug("Checking to see if iDRAC %(host)s is ready",
                      {'host': host})
            if self.lifecycle_service.is_idrac_ready():
                LOG.debug("iDRAC %(host)s is ready", {'host': host})
                return
            LOG.debug("iDRAC %(host)s is not ready", {'host': host})
            retries -= 1
            if retries > 0:
                time.sleep(retry_delay)
        if retries == 0:
            error_msg = ("Timed out waiting iDRAC %(host)s to become "
                         "ready after reset", {'host': host})
            LOG.error(error_msg)
            raise sushy.exceptions.ExtensionError(error=error_msg)

    def _ping_host(self, host):
        """Ping the hostname or IP of a host.

        :param host: Hostname or IP.
        :returns: True if host is alive; otherwise, False.
        """
        response = subprocess.call(["ping", "-c", "1", host])  # noqa:S603,S607
        return response == 0


def get_extension(*args, **kwargs):
    return DellManagerExtension



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values comes from the Redfish Chassis json-schema:
# https://redfish.dmtf.org/schemas/v1/Power.v1_7_1.json

import enum


class PowerSupplyType(enum.Enum):
    UNKNOWN = 'Unknown'
    """The power supply type cannot be determined."""

    AC = 'AC'
    """Alternating Current (AC) power supply."""

    DC = 'DC'
    """Direct Current (DC) power supply."""

    AC_OR_DC = 'ACorDC'
    """The power supply supports both DC or AC."""


# Backward compatibility
POWER_SUPPLY_TYPE_UNKNOWN = PowerSupplyType.UNKNOWN
POWER_SUPPLY_TYPE_AC = PowerSupplyType.AC
POWER_SUPPLY_TYPE_DC = PowerSupplyType.DC
POWER_SUPPLY_TYPE_ACDC = PowerSupplyType.AC_OR_DC


class LineInputVoltageType(enum.Enum):
    UNKNOWN = 'Unknown'
    """The power supply line input voltage type cannot be determined."""

    AC_LOW_LINE = 'ACLowLine'
    """100-127V AC input."""

    AC_MID_LINE = 'ACMidLine'
    """200-240V AC input."""

    AC_HIGH_LINE = 'ACHighLine'
    """277V AC input."""

    DC_NEG48V = 'DCNeg48V'
    """-48V DC input."""

    DC_380V = 'DC380V'
    """High Voltage DC input (380V)."""

    AC_120V = 'AC120V'
    """AC 120V nominal input."""

    AC_240V = 'AC240V'
    """AC 240V nominal input."""

    AC_277V = 'AC277V'
    """AC 277V nominal input."""

    AC_AND_DC_WIDE_RANGE = 'ACandDCWideRange'
    """Wide range AC or DC input."""

    AC_WIDE_RANGE = 'ACWideRange'
    """Wide range AC input."""

    DC_240V = 'DC240V'
    """DC 240V nominal input."""


# Backward compatibility
LINE_INPUT_VOLTAGE_TYPE_UNKNOWN = LineInputVoltageType.UNKNOWN
LINE_INPUT_VOLTAGE_TYPE_ACLOW = LineInputVoltageType.AC_LOW_LINE
LINE_INPUT_VOLTAGE_TYPE_ACMID = LineInputVoltageType.AC_MID_LINE
LINE_INPUT_VOLTAGE_TYPE_ACHIGH = LineInputVoltageType.AC_HIGH_LINE
LINE_INPUT_VOLTAGE_TYPE_DCNEG48 = LineInputVoltageType.DC_NEG48V
LINE_INPUT_VOLTAGE_TYPE_DC380V = LineInputVoltageType.DC_380V
LINE_INPUT_VOLTAGE_TYPE_AC120V = LineInputVoltageType.AC_120V
LINE_INPUT_VOLTAGE_TYPE_AC240V = LineInputVoltageType.AC_240V
LINE_INPUT_VOLTAGE_TYPE_AC277V = LineInputVoltageType.AC_277V
LINE_INPUT_VOLTAGE_TYPE_ACDCWIDE = LineInputVoltageType.AC_AND_DC_WIDE_RANGE
LINE_INPUT_VOLTAGE_TYPE_ACWIDE = LineInputVoltageType.AC_WIDE_RANGE
LINE_INPUT_VOLTAGE_TYPE_DC240V = LineInputVoltageType.DC_240V


class PowerInputType(enum.Enum):
    AC = 'AC'
    """Alternating Current (AC) input range."""

    DC = 'DC'
    """Direct Current (DC) input range."""


# Backward compatibility
INPUT_TYPE_AC = PowerInputType.AC
INPUT_TYPE_DC = PowerInputType.DC



# Copyright (c) 2021-2022 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sushy
from sushy import exceptions
from sushy.oem.dell.resources.system import constants as sys_cons
from sushy.oem.dell.resources.system import raid_service
from sushy.resources.oem import base as oem_base
from sushy import utils as sushy_utils


def _filter_disks_not_in_mode(controller_to_disks, mode):
    """Filters disks that are not in requested mode

    :param controller_to_disks: dictionary of controllers and their drives
    :param mode: constants.PhysicalDiskStateMode
    :returns: dictionary of controllers and their drives that need mode changed
    """
    sushy_raw_device = sushy.VOLUME_TYPE_RAW_DEVICE
    for controller, drives in controller_to_disks.items():
        toprocess_drives = []
        for drive in drives:
            is_raw_device = False
            volumes = None
            try:
                volumes = drive.volumes
            except exceptions.MissingAttributeError:
                pass

            if (volumes
                and (volumes[0].volume_type == sushy_raw_device
                     or volumes[0].raid_type is None)):
                is_raw_device = True

            if (mode == sys_cons.PhysicalDiskStateMode.RAID
                    and is_raw_device
                    or mode == sys_cons.PhysicalDiskStateMode.NONRAID
                    and not is_raw_device):
                toprocess_drives.append(drive)
        controller_to_disks[controller] = toprocess_drives
    return controller_to_disks


class DellSystemExtension(oem_base.OEMResourceBase):

    @property
    @sushy_utils.cache_it
    def raid_service(self):
        """`DellRaidService` of the system"""

        path = sushy_utils.get_sub_resource_path_by(
            self, ["Links", "Oem", "Dell", "DellRaidService"],
            is_collection=False)

        return raid_service.DellRaidService(
            self._conn, path, redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    def change_physical_disk_state(self, mode, controller_to_disks=None):
        """Converts physical disks RAID status

        Converts only those disks that are not already in requested mode.

        :param mode: constants.PhysicalDiskStateMode
        :param controller_to_disks: dictionary of controllers and their drives.
            Optional, if not provided, processes all RAID, except BOSS,
            controller drives.
        :returns: List of task monitors for each controller's disks if any
            drives need changes
        """
        if not controller_to_disks:
            controller_to_disks = self._get_controller_to_disks()

        # Do not process BOSS controllers as can't convert their disks
        boss_controllers = [c for c in controller_to_disks
                            if 'BOSS' in c.name.upper()]
        for c in boss_controllers:
            controller_to_disks.pop(c)

        controller_to_disks = _filter_disks_not_in_mode(
            controller_to_disks, mode)

        # Convert by each controller that have eligible disks
        task_monitors = []
        for controller, drives in controller_to_disks.items():
            if drives:
                drive_fqdds = [d.identity for d in drives]
                if mode == sys_cons.PhysicalDiskStateMode.RAID:
                    task_monitors.append(
                        self.raid_service.convert_to_raid(drive_fqdds))
                elif mode == sys_cons.PhysicalDiskStateMode.NONRAID:
                    task_monitors.append(
                        self.raid_service.convert_to_nonraid(drive_fqdds))

        return task_monitors

    def clear_foreign_config(self, storage_list=None):
        """Clears foreign config on given controllers

        :param storage_list: List of storage objects, each of which
            corresponds to a controller
        :returns: List of task monitors, where each entry is for a
            controller that has foreign config to clear
        """
        if storage_list is None:
            storage_list = self._get_storage_list()

        # Do not process BOSS controllers as not supporting clearing
        boss_storage = [s for s in storage_list
                        if any(c for c in s.storage_controllers
                               if 'BOSS' in c.name.upper())]
        for s in boss_storage:
            storage_list.remove(s)

        task_monitors = []
        for storage in storage_list:
            task_mon = self.raid_service.clear_foreign_config(storage.identity)
            if task_mon:
                task_monitors.append(task_mon)

        return task_monitors

    def _get_controller_to_disks(self):
        """Gets all RAID controllers and their disks on system

        :returns: dictionary of RAID controllers and their disks
        """
        controller_to_disks = {}
        for storage in self._parent_resource.storage.get_members():
            controller = (storage.storage_controllers[0]
                          if storage.storage_controllers else None)
            if not controller or controller and not controller.raid_types:
                continue
            controller_to_disks[controller] = storage.drives
        return controller_to_disks

    def _get_storage_list(self):
        """Gets all storage items corresponding to RAID controllers

        :returns: list of storage items
        """
        storage_list = []
        for storage in self._parent_resource.storage.get_members():
            controller = (storage.storage_controllers[0]
                          if storage.storage_controllers else None)
            if not controller or controller and not controller.raid_types:
                continue
            storage_list.append(storage)
        return storage_list


def get_extension(*args, **kwargs):
    return DellSystemExtension



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from sushy.resources.eventservice import eventdestination
from sushy.tests.unit import base


class EventDestinationTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/eventdestination1.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.subscription = eventdestination.EventDestination(
            self.conn, '/redfish/v1/EventService',
            redfish_version='1.0.0')

    def test__parse_attributes(self):
        self.subscription._parse_attributes(self.json_doc)

        self.assertEqual(self.subscription.identity, '1')
        self.assertEqual(self.subscription.name, 'Event Subscription')
        self.assertEqual(self.subscription.context, 'SomeContext')
        self.assertEqual(self.subscription.destination,
                         'https://localhost/RedfishEvents/EventReceiver.php')
        self.assertEqual(self.subscription.protocol, 'Redfish')
        self.assertEqual(set(self.subscription.event_types),
                         set(["Alert", "StatusChange"]))
        self.assertEqual(self.subscription.http_headers, [])

    def test__delete(self):
        self.subscription.delete()
        self.subscription._conn.delete.assert_called_once_with(
            self.subscription._path)


class EventDestinationCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'eventdestination_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.loads(f.read())
        self.eventdestination = eventdestination.EventDestinationCollection(
            self.conn, '/redfish/v1/EventService/Subscriptions',
            redfish_version='1.0.0')

    @mock.patch.object(eventdestination, 'EventDestination', autospec=True)
    def test_get_member(self, subscription_mock):
        self.eventdestination.get_member(
            '/redfish/v1/EventService/Subscriptions/1')
        subscription_mock.assert_called_once_with(
            self.eventdestination._conn,
            '/redfish/v1/EventService/Subscriptions/1',
            redfish_version=self.eventdestination.redfish_version,
            registries=None, root=self.eventdestination.root)

    @mock.patch.object(eventdestination, 'EventDestination', autospec=True)
    def test_get_members(self, subscription_mock):
        members = self.eventdestination.get_members()
        calls = [
            mock.call(self.eventdestination._conn,
                      f'/redfish/v1/EventService/Subscriptions/{member}',
                      redfish_version=self.eventdestination.redfish_version,
                      registries=None,
                      root=self.eventdestination.root)
            for member in ('1', '2')
        ]
        subscription_mock.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(2, len(members))

    def test__create_subscription_with_required_parameters(self):
        payload = {
            'Protocol': 'Redfish',
            'Destination': 'https://localhost/RedfishEvents/EventReceiver.php',
            'Context': 'IronicContext',
            'EventTypes': ["Alert"]
        }
        with open('sushy/tests/unit/json_samples/eventdestination2.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.conn.post.return_value.status_code = 201
        self.conn.post.return_value.headers.return_value = {
            'Location': '/redfish/v1/EventService/Subscriptions/2'
        }
        new_subscription = self.eventdestination.create(payload)
        self.eventdestination._conn.post.assert_called_once_with(
            '/redfish/v1/EventService/Subscriptions',
            data=payload,
        )
        self.assertIsNotNone(new_subscription)
        self.assertEqual(new_subscription.identity, '2')
        self.assertEqual(new_subscription.context, 'IronicContext')
        self.assertEqual(new_subscription.destination,
                         'https://localhost/RedfishEvents/EventReceiver.php')
        self.assertIn("Alert", new_subscription.event_types)

    def test__create_subscription_with_invalid_event_type(self):
        payload = {
            'Protocol': 'Redfish',
            'Destination': 'https://localhost/RedfishEvents/EventReceiver.php',
            'Context': 'IronicContext',
            'EventTypes': ["Other"]
        }
        self.conn.post.return_value.status_code = 400
        new_subscription = self.eventdestination.create(payload)
        self.eventdestination._conn.post.assert_called_once_with(
            '/redfish/v1/EventService/Subscriptions',
            data=payload,
        )
        self.assertIsNone(new_subscription)

    def test__create_subscription_with_empty_context(self):
        payload = {
            'Protocol': 'Redfish',
            'Destination': 'https://localhost/RedfishEvents/EventReceiver.php',
            'Context': '',
            'EventTypes': ["Alert"]
        }
        with open('sushy/tests/unit/json_samples/eventdestination3.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.conn.post.return_value.status_code = 201
        self.conn.post.return_value.headers.return_value = {
            'Location': '/redfish/v1/EventService/Subscriptions/3'
        }
        new_subscription = self.eventdestination.create(payload)
        self.eventdestination._conn.post.assert_called_once_with(
            '/redfish/v1/EventService/Subscriptions',
            data=payload,
        )
        self.assertIsNotNone(new_subscription)
        self.assertEqual(new_subscription.identity, '3')
        self.assertIsNone(new_subscription.context)
        self.assertEqual(new_subscription.destination,
                         'https://localhost/RedfishEvents/EventReceiver.php')
        self.assertIn("Alert", new_subscription.event_types)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/Certificate.v1_4_0.json
# https://redfish.dmtf.org/schemas/v1/CertificateCollection.json

from dateutil import parser

from sushy.resources import base
from sushy.resources.certificateservice import constants as cert_cons


class Identifier(base.CompositeField):
    """The identifier information about a certificate."""

    city = base.Field('City')
    common_name = base.Field('CommonName')
    country = base.Field('Country')
    email = base.Field('Email')
    organization = base.Field('Organization')
    organizational_unit = base.Field('OrganizationalUnit')
    state = base.Field('State')


class Certificate(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The certificate identity string"""

    name = base.Field('Name')
    """The certificate name"""

    certificate_string = base.Field('CertificateString')
    """Certificate in the format defined by certificate_type"""

    certificate_type = base.MappedField('CertificateType',
                                        cert_cons.CertificateType)
    """The format of the certificate"""

    certificate_usage_type = base.MappedField('CertificateUsageType',
                                              cert_cons.CertificateUsageType)
    """The types or purposes for this certificate"""

    description = base.Field('Description')
    """Certificate description"""

    fingerprint = base.Field('Fingerprint')
    """The fingerprint of the certificate"""

    fingerprint_hash_algorithm = base.Field('FingerprintHashAlgorithm')
    """The hash algorithm for the fingerprint of the certificate"""

    issuer = Identifier('Issuer')
    """The issuer of the certificate"""

    key_usage = base.MappedListField('KeyUsage', cert_cons.KeyUsage)
    """The key usage extension, which defines the purpose of the public keys
    in this certificate"""

    serial_number = base.Field('SerialNumber')
    """The serial number of the certificate"""

    signature_algorithm = base.Field('SignatureAlgorithm')
    """The algorithm used for creating the signature of the certificate"""

    subject = Identifier('Subject')
    """The subject of the certificate"""

    uefi_signature_owner = base.Field('UefiSignatureOwner')
    """The UEFI signature owner for this certificate"""

    valid_not_after = base.Field('ValidNotAfter', adapter=parser.parse)
    """The date when the certificate is no longer valid"""

    valid_not_before = base.Field('ValidNotBefore', adapter=parser.parse)
    """The date when the certificate becomes valid"""

    # TODO(dtantsur): actions

    def delete(self):
        """Delete this certificate."""
        self._conn.delete(self._path)


# Yes, certificate collection is not the same thing as certificate locations.
# The latter is only used in CertificateService, while the former - in every
# place where certificates are supported, e.g. virtual media. For this reason
# there is no link from CertificateService to CertificateCollection.
class CertificateCollection(base.MutableResourceCollectionBase):

    _resource_type = Certificate

    def create_member(self, certificate_string, certificate_type):
        """Create a new member of this collection.

        :param certificate_string: the contents of the new certificate.
        :param certificate_type: the type of the new certificate, one of
            :py:class:`sushy.CertificateType`.
        """
        return self._create_member(dict(
            CertificateString=certificate_string,
            CertificateType=cert_cons.CertificateType(certificate_type).value,
        ))



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import logging


LOG = logging.getLogger(__name__)


class SushyError(Exception):
    """Basic exception for errors raised by Sushy"""

    message = None

    def __init__(self, message=None, **kwargs):
        if message is not None:
            self.message = message
        if self.message and kwargs:
            self.message = self.message % kwargs

        super().__init__(self.message)


class ConnectionError(SushyError):
    message = 'Unable to connect to %(url)s. Error: %(error)s'


class MissingAttributeError(SushyError):
    message = ('The attribute %(attribute)s is missing from the '
               'resource %(resource)s')


class MalformedAttributeError(SushyError):
    message = ('The attribute %(attribute)s is malformed in the '
               'resource %(resource)s: %(error)s')


class MissingActionError(SushyError):
    message = ('The action %(action)s is missing from the '
               'resource %(resource)s')


class InvalidParameterValueError(SushyError):
    message = ('The parameter "%(parameter)s" value "%(value)s" is invalid. '
               'Valid values are: %(valid_values)s')


class ArchiveParsingError(SushyError):
    message = 'Failed parsing archive "%(path)s": %(error)s'


class UnknownDefaultError(SushyError):
    message = 'Failed at determining default for "%(entity)s": %(error)s'


class ExtensionError(SushyError):
    message = ('Sushy Extension Error: %(error)s')


class OEMExtensionNotFoundError(SushyError):
    message = 'No %(resource)s OEM extension found by name "%(name)s".'


class MissingHeaderError(SushyError):
    message = 'Response to %(target_uri)s did not contain a %(header)s header'


class HTTPError(SushyError):
    """Basic exception for HTTP errors"""

    status_code = None
    """HTTP status code."""

    body = None
    """Error JSON body, if present."""

    code = 'Base.1.0.GeneralError'
    """Error code defined in the Redfish specification, if present."""

    detail = None
    """Error message defined in the Redfish specification, if present."""

    message = ('HTTP %(method)s %(url)s returned code %(code)s. %(error)s '
               'Extended information: %(ext_info)s')

    extended_info = None
    """Extended information provided in the response."""

    def __init__(self, method, url, response):
        self.status_code = response.status_code
        try:
            body = response.json()
        except ValueError:
            LOG.warning('Error response from %(method)s %(url)s '
                        'with status code %(code)s has no JSON body',
                        {'method': method, 'url': url, 'code':
                         self.status_code})
            error = 'unknown error'
        else:
            self.body = body.get('error', {})
            self.code = self.body.get('code', 'Base.1.0.GeneralError')
            self.detail = self.body.get('message')
            self.extended_info = self.body.get('@Message.ExtendedInfo')
            message = self._get_most_severe_msg(self.extended_info or [{}])
            self.detail = message or self.detail
            error = '{}: {}'.format(self.code, self.detail or 'unknown error.')
        kwargs = {'method': method, 'url': url, 'code': self.status_code,
                  'error': error, 'ext_info': self.extended_info}
        LOG.debug('HTTP response for %(method)s %(url)s: '
                  'status code: %(code)s, error: %(error)s, '
                  'extended: %(ext_info)s', kwargs)
        super().__init__(**kwargs)

    @staticmethod
    def _get_most_severe_msg(extended_info):
        if not isinstance(extended_info, list):
            return extended_info.get('Message', None)
        if len(extended_info) > 0:
            for sev in ['Critical', 'Warning']:
                for i, m in enumerate(extended_info):
                    if m.get('Severity') == sev:
                        return m.get('Message')

    @property
    def related_properties(self):
        """List of properties related to the error."""
        try:
            return self.extended_info[0]['RelatedProperties']
        except (IndexError, KeyError, TypeError):
            return []


class BadRequestError(HTTPError):
    pass


class ResourceNotFoundError(HTTPError):
    # Overwrite the complex generic message with a simpler one.
    message = 'Resource %(url)s not found'


class ServerSideError(HTTPError):
    pass


class AccessError(HTTPError):
    pass


class NotAcceptableError(HTTPError):
    pass


class MissingXAuthToken(HTTPError):
    message = ('No X-Auth-Token returned from remote host when '
               'attempting to establish a session. Error: %(error)s')


def raise_for_response(method, url, response):
    """Raise a correct error class, if needed."""
    if response.status_code < http_client.BAD_REQUEST:
        return
    elif response.status_code == http_client.NOT_FOUND:
        raise ResourceNotFoundError(method, url, response)
    elif response.status_code == http_client.BAD_REQUEST:
        raise BadRequestError(method, url, response)
    elif response.status_code in (http_client.UNAUTHORIZED,
                                  http_client.FORBIDDEN):
        raise AccessError(method, url, response)
    elif response.status_code == http_client.NOT_ACCEPTABLE:
        raise NotAcceptableError(method, url, response)
    elif response.status_code >= http_client.INTERNAL_SERVER_ERROR:
        raise ServerSideError(method, url, response)
    else:
        raise HTTPError(method, url, response)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy import main


class RootTestCase(BaseTestCase):

    @mock.patch('sushy.auth.SessionOrBasicAuth', autospec=True)
    @mock.patch('sushy.connector.Connector', autospec=True)
    @mock.patch('sushy.resources.sessionservice.sessionservice.'
                'SessionService', autospec=True)
    def setUp(self, mock_session_service, mock_connector, mock_auth):
        super().setUp()
        self.conn = mock.Mock()
        self.sess_serv = mock.Mock()
        self.sess_serv.create_session.return_value = (None, None)
        mock_session_service.return_value = self.sess_serv
        mock_connector.return_value = self.conn
        with open('sushy/tests/oem/dell/unit/json_samples/root.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.root = main.Sushy('http://foo.bar:1234',
                               verify=True, auth=mock_auth)

    def test_oem_vendors(self):
        self.assertEqual(['Dell'], self.root.oem_vendors)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/EventDestination.v1_0_0.json

import logging

from sushy.resources import base

LOG = logging.getLogger(__name__)


class EventDestination(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The EventDestination resource identity"""

    name = base.Field('Name', required=True)
    """The EventDestination resource name"""

    context = base.Field('Context')
    """A client-supplied string that is stored with the event destination
    subscription"""

    description = base.Field('Description')
    """The description of the EventDestination resource"""

    destination = base.Field('Destination')
    """The URI of the destination Event Service"""

    event_types = base.Field('EventTypes', adapter=list)
    """The types of events that shall be sent to the destination"""

    protocol = base.Field('Protocol')
    """Contain the protocol type that the event will use for sending
    the event to the destination.  A value of Redfish shall be used
    to indicate that the event type shall adhere to that defined in
    the Redfish specification"""

    http_headers = base.Field('HttpHeaders', adapter=list)
    """This is for setting HTTP headers, such as authorization information.
    This object will be null on a GET."""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing an EventDestination

        :param connector: A Connector instance
        :param identity: The identity of the EventDestination resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of registries to be used in any resource
            that needs registries to parse messages.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)

    def delete(self):
        """Delete an EventDestination

        :raises: ConnectionError
        :raises: HTTPError
        """
        self._conn.delete(self._path)


class EventDestinationCollection(base.ResourceCollectionBase):

    name = base.Field('Name')
    """The EventDestination collection name"""

    description = base.Field('Description')
    """The EventDestination collection description"""

    @property
    def _resource_type(self):
        return EventDestination

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a EventDestinationCollection

        :param connector: A Connector instance
        :param identity: The identity of the EventDestination resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of registries to be used in any resource
            that needs registries to parse messages.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)

    def _create(self, payload):
        r = self._conn.post(self._path, data=payload)
        location = r.headers.get('Location')
        return r, location

    def create(self, payload):
        """Create a Subscription

        :param payload: The payload representing the subscription.

        :raises: ConnectionError
        :raises: HTTPError
        :returns: The new subscription
        """
        r, location = self._create(payload)
        if r.status_code == 201:
            if location:
                self.refresh()
                return self.get_member(location)



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy.oem.dell.resources.manager import job_service


class DellJobServiceTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'job_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers.get.return_value = '1'
        self.job_service = job_service.DellJobService(
            self.conn,
            '/redfish/v1/Dell/Managers/iDRAC.Embedded.1/DellJobService')

    def test_delete_jobs(self):
        job_ids = ['JID_471269252011', 'JID_471269252012']
        self.job_service.delete_jobs(job_ids=job_ids)
        target_uri = ('/redfish/v1/Dell/Managers'
                      '/iDRAC.Embedded.1/DellJobService'
                      '/Actions/''DellJobService.DeleteJobQueue')
        for job_id in job_ids:
            self.conn.post.assert_any_call(target_uri,
                                           data={'JobID': job_id})
        self.assertEqual(2, self.conn.post.call_count)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from sushy.resources import constants as res_cons
from sushy.resources.system import constants as sys_cons
from sushy.resources.system import pcie_device
from sushy.tests.unit import base


class PCIeDeviceTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/pcie_device.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.pcie_dev = pcie_device.PCIeDevice(
            self.conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.pcie_dev._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.pcie_dev.redfish_version)
        self.assertEqual('196-0', self.pcie_dev.identity)
        self.assertEqual('BCM957414A4142CC 10Gb/25Gb Ethernet PCIe',
                         self.pcie_dev.name)
        self.assertEqual('BCM957414A4142CC 10Gb/25Gb Ethernet PCIe',
                         self.pcie_dev.description)
        self.assertEqual('Broadcom Inc. and subsidiaries',
                         self.pcie_dev.manufacturer)
        self.assertEqual(sys_cons.PCIeDeviceType.MULTI_FUNCTION,
                         self.pcie_dev.device_type)
        self.assertEqual('224.1.102.0', self.pcie_dev.firmware_version)
        self.assertIsNone(self.pcie_dev.asset_tag)
        self.assertIsNone(self.pcie_dev.serial_number)
        self.assertIsNone(self.pcie_dev.model)
        self.assertIsNone(self.pcie_dev.part_number)
        self.assertIsNone(self.pcie_dev.sku)

    def test_pcie_interface(self):
        pcie_if = self.pcie_dev.pcie_interface
        self.assertEqual('Gen5', pcie_if.pcie_type)
        self.assertEqual('Gen5', pcie_if.max_pcie_type)
        self.assertEqual(8, pcie_if.lanes_in_use)
        self.assertEqual(8, pcie_if.max_lanes)

    def test_slot_information(self):
        slot = self.pcie_dev.slot
        self.assertEqual(sys_cons.SlotType.HALF_LENGTH, slot.slot_type)
        self.assertEqual('Gen5', slot.pcie_type)
        self.assertEqual(8, slot.lanes)
        self.assertIsNotNone(slot.location)
        self.assertIsNone(slot.lane_splitting)
        self.assertIsNone(slot.hot_pluggable)

    def test_status_field(self):
        status = self.pcie_dev.status
        self.assertEqual(res_cons.State.ENABLED, status.state)
        self.assertEqual(res_cons.Health.OK, status.health)
        self.assertEqual(res_cons.Health.OK, status.health_rollup)

    def test_links_field(self):
        links = self.pcie_dev.links
        self.assertIsNotNone(links)


class PCIeDeviceOnboardTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'pcie_device_onboard.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.onboard_dev = pcie_device.PCIeDevice(
            self.conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/128-1',
            redfish_version='1.0.2')

    def test_onboard_device_attributes(self):
        self.assertEqual('128-1', self.onboard_dev.identity)
        self.assertEqual('Advanced Micro Devices, Inc. [AMD]',
                         self.onboard_dev.name)
        self.assertEqual('Advanced Micro Devices, Inc. [AMD]',
                         self.onboard_dev.manufacturer)
        self.assertEqual(sys_cons.PCIeDeviceType.MULTI_FUNCTION,
                         self.onboard_dev.device_type)
        self.assertEqual('', self.onboard_dev.firmware_version)

    def test_onboard_device_empty_slot(self):
        slot = self.onboard_dev.slot
        self.assertIsNotNone(slot)
        # Empty slot object should have None values
        self.assertIsNone(slot.slot_type)
        self.assertIsNone(slot.lanes)
        self.assertIsNone(slot.pcie_type)


class PCIeDeviceCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'pcie_device_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.collection = pcie_device.PCIeDeviceCollection(
            self.conn, '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.collection._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.collection.redfish_version)
        self.assertEqual('PCIe Device Collection', self.collection.name)
        self.assertEqual(
            ('/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0',
             '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/65-0'),
            self.collection.members_identities)

    @mock.patch.object(pcie_device, 'PCIeDevice', autospec=True)
    def test_get_member(self, mock_pcie_device):
        self.collection.get_member(
            '/redfish/v1/Chassis/System.Embedded.1/'
            'PCIeDevices/196-0')
        mock_pcie_device.assert_called_once_with(
            self.collection._conn,
            '/redfish/v1/Chassis/System.Embedded.1/'
            'PCIeDevices/196-0',
            redfish_version=self.collection.redfish_version,
            registries=None, root=self.collection.root)

    @mock.patch.object(pcie_device, 'PCIeDevice', autospec=True)
    def test_get_members(self, mock_pcie_device):
        members = self.collection.get_members()
        calls = [
            mock.call(self.collection._conn,
                      '/redfish/v1/Chassis/System.Embedded.1/'
                      'PCIeDevices/196-0',
                      redfish_version=self.collection.redfish_version,
                      registries=None, root=self.collection.root),
            mock.call(self.collection._conn,
                      '/redfish/v1/Chassis/System.Embedded.1/'
                      'PCIeDevices/65-0',
                      redfish_version=self.collection.redfish_version,
                      registries=None, root=self.collection.root)
        ]
        mock_pcie_device.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(2, len(members))


class PCIeFunctionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/pcie_function.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.pcie_func = pcie_device.PCIeFunction(
            self.conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
            'PCIeFunctions/196-0-1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.pcie_func._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.pcie_func.redfish_version)
        self.assertEqual('196-0-1', self.pcie_func.identity)
        self.assertEqual('BCM957414A4142CC 10Gb/25Gb Ethernet PCIe',
                         self.pcie_func.name)
        self.assertEqual('BCM957414A4142CC 10Gb/25Gb Ethernet PCIe',
                         self.pcie_func.description)

    def test_function_properties(self):
        self.assertEqual(1, self.pcie_func.function_id)
        self.assertEqual(sys_cons.FunctionType.PHYSICAL,
                         self.pcie_func.function_type)
        self.assertEqual(sys_cons.FunctionProtocol.PCIE,
                         self.pcie_func.function_protocol)
        self.assertEqual(sys_cons.DeviceClass.NETWORK_CONTROLLER,
                         self.pcie_func.device_class)
        self.assertTrue(self.pcie_func.enabled)

    def test_device_identification_fields(self):
        self.assertEqual('0x16d7', self.pcie_func.device_id)
        self.assertEqual('0x14e4', self.pcie_func.vendor_id)
        self.assertEqual('0x1402', self.pcie_func.subsystem_id)
        self.assertEqual('0x14e4', self.pcie_func.subsystem_vendor_id)
        self.assertEqual('0x01', self.pcie_func.revision_id)
        self.assertEqual('0x000002', self.pcie_func.class_code)

    def test_pci_addressing_fields(self):
        self.assertEqual('0x03', self.pcie_func.bus_number)
        self.assertEqual('0x00', self.pcie_func.device_number)
        self.assertEqual('0x01', self.pcie_func.function_number)
        self.assertEqual('0x0000', self.pcie_func.segment_number)

    def test_status_field(self):
        status = self.pcie_func.status
        self.assertEqual(res_cons.State.ENABLED, status.state)
        self.assertEqual(res_cons.Health.OK, status.health)
        self.assertEqual(res_cons.Health.OK, status.health_rollup)

    def test_links_field(self):
        links = self.pcie_func.links
        self.assertIsNotNone(links)


class PCIeFunctionCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'pcie_function_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.collection = pcie_device.PCIeFunctionCollection(
            self.conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
            'PCIeFunctions',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.collection._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.collection.redfish_version)
        self.assertEqual('PCIeFunction Collection', self.collection.name)
        self.assertEqual(
            ('/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
             'PCIeFunctions/196-0-0',
             '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
             'PCIeFunctions/196-0-1'),
            self.collection.members_identities)

    def test_function_count_property(self):
        with mock.patch.object(
            self.collection, 'get_members', autospec=True,
            return_value=[mock.Mock(), mock.Mock()]):
            self.assertEqual(2, self.collection.function_count)

    @mock.patch.object(pcie_device, 'PCIeFunction', autospec=True)
    def test_get_member(self, mock_pcie_function):
        self.collection.get_member(
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
            'PCIeFunctions/196-0-1')
        mock_pcie_function.assert_called_once_with(
            self.collection._conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
            'PCIeFunctions/196-0-1',
            redfish_version=self.collection.redfish_version,
            registries=None, root=self.collection.root)

    @mock.patch.object(pcie_device, 'PCIeFunction', autospec=True)
    def test_get_members(self, mock_pcie_function):
        members = self.collection.get_members()
        calls = [
            mock.call(self.collection._conn,
                      '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/'
                      '196-0/PCIeFunctions/196-0-0',
                      redfish_version=self.collection.redfish_version,
                      registries=None, root=self.collection.root),
            mock.call(self.collection._conn,
                      '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/'
                      '196-0/PCIeFunctions/196-0-1',
                      redfish_version=self.collection.redfish_version,
                      registries=None, root=self.collection.root)
        ]
        mock_pcie_function.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(2, len(members))


class PCIeDeviceFunctionsTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/pcie_device.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.pcie_dev = pcie_device.PCIeDevice(
            self.conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0',
            redfish_version='1.0.2')

    @mock.patch.object(pcie_device, 'PCIeFunctionCollection', autospec=True)
    def test_pcie_functions_property(self, mock_collection):
        # Test normal case with PCIeFunctions present
        _ = self.pcie_dev.pcie_functions
        mock_collection.assert_called_once_with(
            self.pcie_dev._conn,
            '/redfish/v1/Chassis/System.Embedded.1/PCIeDevices/196-0/'
            'PCIeFunctions',
            redfish_version=self.pcie_dev.redfish_version,
            registries=self.pcie_dev.registries,
            root=self.pcie_dev.root)

    @mock.patch.object(pcie_device, 'PCIeFunctionCollection', autospec=True)
    @mock.patch('sushy.utils.get_sub_resource_path_by', autospec=True)
    def test_pcie_functions_missing_attribute(
        self, mock_get_path, mock_collection):
        from sushy import exceptions
        # Simulate MissingAttributeError when PCIeFunctions is not supported
        mock_get_path.side_effect = exceptions.MissingAttributeError(
            attribute='PCIeFunctions', resource=self.pcie_dev._path)

        _ = self.pcie_dev.pcie_functions
        # Should create empty collection

        mock_collection.assert_called_once_with(
            self.pcie_dev._conn, "/empty",
            redfish_version=self.pcie_dev.redfish_version,
            registries=self.pcie_dev.registries,
            root=self.pcie_dev.root)



# Copyright (c) 2022 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum


class ControllerMode(enum.Enum):
    """RAID controller modes."""

    RAID = "RAID"
    """RAID mode."""

    HBA = "HBA"
    """HBA/Passthru mode. Does not support RAID. For PERC 9 controllers."""

    EHBA = "EnhancedHBA"
    """Enhanced HBA mode. Limited RAID support. For PERC 10 controllers."""



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources.system import constants

LOG = logging.getLogger(__name__)


class ResetKeysActionField(common.ActionField):

    allowed_values = base.Field('ResetKeysType@Redfish.AllowableValues',
                                adapter=list)


class ActionsField(base.CompositeField):

    reset_keys = ResetKeysActionField('#SecureBootDatabase.ResetKeys')
    """Action that resets the UEFI Secure Boot keys."""


class SecureBootDatabase(base.ResourceBase):

    # TODO(dtantsur): certificates

    database_id = base.MappedField('DatabaseId',
                                   constants.SecureBootDatabaseId)
    """Standard UEFI database type."""

    description = base.Field('Description')
    """The system description"""

    identity = base.Field('Id', required=True)
    """The secure boot database identity string"""

    name = base.Field('Name')
    """The secure boot database name"""

    # TODO(dtantsur): signatures

    _actions = ActionsField('Actions')

    def _get_reset_action_element(self):
        reset_action = self._actions.reset_keys
        if not reset_action:
            raise exceptions.MissingActionError(
                action='#SecureBootDatabase.ResetKeys', resource=self._path)
        return reset_action

    def get_allowed_reset_keys_values(self):
        """Get the allowed values for resetting the keys.

        :returns: A set with the allowed values.
        """
        reset_action = self._get_reset_action_element()

        if not reset_action.allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'reset keys action for %s', self.identity)
            return constants._SECURE_BOOT_DATABASE_RESET_KEYS

        return {v for v in constants._SECURE_BOOT_DATABASE_RESET_KEYS
                if v.value in reset_action.allowed_values}

    def reset_keys(self, reset_type):
        """Reset secure boot keys.

        :param reset_type: Reset type, one of `SECURE_BOOT_RESET_KEYS_*`
            constants.
        """
        valid_resets = self.get_allowed_reset_keys_values()
        if reset_type not in valid_resets:
            raise exceptions.InvalidParameterValueError(
                parameter='reset_type', value=reset_type,
                valid_values=valid_resets)

        reset_type = constants.SecureBootResetKeysType(reset_type).value
        target_uri = self._get_reset_action_element().target_uri
        self._conn.post(target_uri, data={'ResetKeysType': reset_type})


class SecureBootDatabaseCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return SecureBootDatabase

    def __init__(self, connector, path, redfish_version=None, registries=None,
                 root=None):
        """A class representing a ComputerSystemCollection

        :param connector: A Connector instance
        :param path: The canonical path to the System collection resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, path, redfish_version=redfish_version,
            registries=registries, root=root)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import os
import sys

import sushy
from sushy.oem.dell import utils

USERNAME = 'root'
PASSWORD = 'calvin'

SERVICE_ROOT = 'http://demo.snmplabs.com:80/redfish/v1'

SYSTEM_ID = '437XR1138R2'

BOOT_DEVICE = sushy.VIRTUAL_MEDIA_CD
BOOT_MODE = sushy.BOOT_SOURCE_MODE_BIOS

BOOT_IMAGE = 'http://demo.snmplabs.com/mini.iso'

LOG = logging.getLogger(__name__)


def main():
    """Boot Dell node from virtual media device"""

    LOG.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    LOG.addHandler(handler)

    authenticator = sushy.auth.BasicAuth(USERNAME, PASSWORD)

    conn = sushy.Sushy(SERVICE_ROOT, verify=False, auth=authenticator)

    LOG.info('connected to %s', SERVICE_ROOT)

    system = conn.get_system(
        os.path.join(SERVICE_ROOT, 'Systems', SYSTEM_ID))

    LOG.info('read system resource %s', system.identity)

    for manager in system.managers:

        LOG.info('trying manager %s', manager.identity)

        for v_media in manager.virtual_media.get_members():
            if BOOT_DEVICE not in v_media.media_types:
                continue

            LOG.info(
                'device %s is present at %s', BOOT_DEVICE, manager.identity)

            try:
                manager_oem = manager.get_oem_extension('Dell')

            except sushy.exceptions.OEMExtensionNotFoundError:
                LOG.info('Dell OEM not found')
                continue

            LOG.info('found Dell OEM extension at %s', manager.identity)

            if v_media.inserted:
                v_media.eject_media()

                LOG.info('ejected virtual media')

            v_media.insert_media(BOOT_IMAGE, inserted=True,
                                 write_protected=True)

            LOG.info('inserted boot image %s into virtual media', BOOT_IMAGE)

            # the caller (e.g. ironic) sets boot mode first, boot device second
            system.set_system_boot_source(
                BOOT_DEVICE, enabled=sushy.BOOT_SOURCE_ENABLED_CONTINUOUS,
                mode=BOOT_MODE)

            # with Dell, patching System tree does not work as expected
            # we need to reboot for the new boot mode to take effect
            utils.reboot_system(system)

            LOG.info('set boot mode to %s', BOOT_MODE)

            manager_oem.set_virtual_boot_device(
                BOOT_DEVICE, persistent=False, manager=manager, system=system)

            LOG.info('set boot device to %s', BOOT_DEVICE)

            # real caller should better not use our way to reboot
            utils.reboot_system(system)

            LOG.info('system rebooted')

            return 0


if __name__ == '__main__':
    sys.exit(main())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/SoftwareInventory.v1_2_0.json

import logging

from sushy.resources import base
from sushy.resources import common

LOG = logging.getLogger(__name__)


class SoftwareInventory(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The software inventory identity"""

    lowest_supported_version = base.Field('LowestSupportedVersion')
    """The lowest supported version of the software"""

    manufacturer = base.Field('Manufacturer')
    """The manufacturer of the software"""

    name = base.Field('Name', required=True)
    """The software inventory name"""

    release_date = base.Field('ReleaseDate')
    """Release date of the software"""

    related_item = base.Field('RelatedItem')
    """The ID(s) of the resources associated with the software inventory
    item"""

    status = common.StatusField('Status')
    """The status of the software inventory"""

    software_id = base.Field('SoftwareId')
    """The identity of the software"""

    uefi_device_paths = base.Field('UefiDevicePaths')
    """Represents the UEFI Device Path(s)"""

    updateable = base.Field('Updateable')
    """Indicates whether this software can be updated by the update
    service"""

    version = base.Field('Version')
    """The version of the software"""

    def __init__(self, connector, identity,
                 redfish_version=None, registries=None, root=None):
        """A class representing a SoftwareInventory

        :param connector: A Connector instance
        :param identity: The identity of the SoftwareInventory resources
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)


class SoftwareInventoryCollection(base.ResourceCollectionBase):

    name = base.Field('Name')
    """The software inventory collection name"""

    description = base.Field('Description')
    """The software inventory collection description"""

    @property
    def _resource_type(self):
        return SoftwareInventory

    def __init__(self, connector, identity,
                 redfish_version=None, registries=None, root=None):
        """A class representing a SoftwareInventoryCollection

        :param connector: A Connector instance
        :param identity: The identity of SoftwareInventory resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/VirtualMedia.v1_2_0.json

from http import client as http_client
import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources.certificateservice import certificate
from sushy.resources import common
from sushy.resources.manager import constants as mgr_cons
from sushy import utils

LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):

    insert_media = common.ActionField("#VirtualMedia.InsertMedia")
    eject_media = common.ActionField("#VirtualMedia.EjectMedia")


class VirtualMedia(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """Virtual Media resource identity string"""

    name = base.Field('Name', required=True)
    """The name of resource"""

    connected_via = base.MappedField('ConnectedVia', mgr_cons.ConnectedVia)
    """Current virtual media connection methods

    Applet: Connected to a client application
    NotConnected: No current connection
    Oem: Connected via an OEM-defined method
    URI: Connected to a URI location
    """

    image = base.Field('Image')
    """A URI providing the location of the selected image"""

    image_name = base.Field('ImageName')
    """The image name"""

    inserted = base.Field('Inserted')
    """Indicates if virtual media is inserted in the virtual device"""

    media_types = base.MappedListField(
        'MediaTypes', mgr_cons.VirtualMediaType, default=[])
    """List of supported media types as virtual media"""

    status = common.StatusField('Status')
    """The virtual media status"""

    transfer_method = base.MappedField('TransferMethod',
                                       mgr_cons.TransferMethod)
    """The transfer method to use with the Image"""

    user_name = base.Field('UserName')
    """The user name to access the Image parameter-specified URI"""

    verify_certificate = base.Field('VerifyCertificate', adapter=bool)
    """Whether to verify the certificate of the server for the Image"""

    write_protected = base.Field('WriteProtected')
    """Indicates the media is write protected"""

    _actions = ActionsField('Actions')
    """Insert/eject action for virtual media"""

    _certificates_path = base.Field(['Certificates', '@odata.id'])

    def _get_insert_media_uri(self):
        insert_media = self._actions.insert_media if self._actions else None
        use_patch = False
        if not insert_media:
            insert_uri = self.path
            use_patch = self._allow_patch()
            if not use_patch:
                raise exceptions.MissingActionError(
                    action='#VirtualMedia.InsertMedia', resource=self._path)
        else:
            insert_uri = insert_media.target_uri
        return insert_uri, use_patch

    def _get_eject_media_uri(self):
        eject_media = self._actions.eject_media if self._actions else None
        use_patch = False
        if not eject_media:
            eject_uri = self.path
            use_patch = self._allow_patch()
            if not use_patch:
                raise exceptions.MissingActionError(
                    action='#VirtualMedia.EjectMedia', resource=self._path)
        else:
            eject_uri = eject_media.target_uri
        return eject_uri, use_patch

    def is_transfer_protocol_required(self, error=None):
        """Check the response code and body and in case of failure

        Try to determine if it happened due to missing TransferProtocolType.
        """
        if (error.code.endswith('GeneralError')
           and 'TransferProtocolType' in error.detail):
            return True

        return (
            (error.code.endswith(".ActionParameterMissing")
             or error.code.endswith(".PropertyMissing"))
            and (("#/TransferProtocolType" in error.related_properties)
                 or ("/TransferProtocolType" in error.related_properties))
        )

    def is_transfer_method_required(self, error=None):
        """Check the response code and body and in case of failure

        Try to determine if it happened due to missing TransferMethod
        """
        if (error.code.endswith('GeneralError')
                and 'TransferMethod' in error.detail):
            return True
        return False

    def is_credentials_required(self, error=None):
        """Check the response code and body and in case of failure

        Try to determine if it happened due to missing Credentials
        """

        # Accept broader variants seen on NVIDIA DGX where "code" may be absent
        try:
            body = error.response.json()
        except Exception:
            body = {}

        # Structured path: @Message.ExtendedInfo entries
        ext = (body.get('error', {}) or {}).get('@Message.ExtendedInfo',
                                                []) or []
        for m in ext:
            mid = (m.get('MessageId') or '')
            msg = (m.get('Message') or '')
            if mid.endswith('ActionParameterMissing') and (
                'UserName' in msg or 'Password' in msg):
                return True

        # Legacy/structured: error.code + message mentions UserName/Password
        err_obj = (body.get('error') or {})
        code = (err_obj.get('code') or '')
        msg = (err_obj.get('message') or '')
        if code.endswith('GeneralError') and (
            'UserName' in msg or 'Password' in msg
                or 'parameter missing' in msg):
            return True

        # Text fallback: common strings observed in the field
        raw_msg = ((body.get('error') or {}).get('message') or '') + ' ' + (
            error.detail or '')
        if ('requires the parameter UserName' in raw_msg
                or 'requires the parameter Password' in raw_msg):
            return True
        return False

    def insert_media(self, image, inserted=True, write_protected=True,
                     username=None, password=None, transfer_method=None):
        """Attach remote media to virtual media

        :param image: a URI providing the location of the selected image
        :param inserted: specify if the image is to be treated as inserted upon
            completion of the action.
        :param write_protected: indicates the media is write protected
        :param username: User name for the image URI.
        :param password: Password for the image URI.
        :param transfer_method: Transfer method (stream or upload) to use
            for the image.
        """
        target_uri, use_patch = self._get_insert_media_uri()
        # NOTE(janders) Inserted and WriteProtected attributes are optional
        # as per Redfish schema 2021.1. However - some BMCs (e.g. Lenovo SD530
        # which is using PATCH method as opposed to InsertMedia action) will
        # not attach vMedia if Inserted is not specified.
        # On the other hand, machines such as SuperMicro X11 will return
        # an error if Inserted or WriteProtected are specified. In order to
        # make both work, we remove Inserted and WriteProtected from payload
        # for BMCs which don't use PATCH if their values are set to defaults
        # as per the spec (True, True). We continue to set Inserted and
        # WriteProtected in payload if PATCH method is used.
        payload = {'Image': image}
        if username is not None:
            payload['UserName'] = username
        if password is not None:
            payload['Password'] = password
        if transfer_method is not None:
            try:
                payload['TransferMethod'] = \
                    mgr_cons.TransferMethod(transfer_method).value
            except ValueError:
                raise exceptions.InvalidParameterValueError(
                    parameter='transfer_method',
                    value=transfer_method,
                    valid_values=', '.join(map(str, mgr_cons.TransferMethod)))

        if use_patch:
            payload['Inserted'] = inserted
            payload['WriteProtected'] = write_protected
            headers = None
            etag = self._get_etag()
            if etag is not None:
                headers = {"If-Match": etag}
            self._conn.patch(target_uri, data=payload, headers=headers)
        else:
            # NOTE(janders) only include Inserted and WriteProtected
            # in request payload if values other than defaults (True,True)
            # are set (fix for SuperMicro X11/X12).
            if not inserted:
                payload['Inserted'] = False
            if not write_protected:
                payload['WriteProtected'] = False
            # Called a Max of 4 times, 1 plain + 1 for each of the 3 errors
            for _ in range(4):
                try:
                    self._conn.post(target_uri, data=payload)
                    break
                except exceptions.HTTPError as error:
                    # NOTE(janders) attempting to detect whether attachment
                    # failure is due to absence of TransferProtocolType param
                    # and if so adding it
                    if payload.get('TransferProtocolType') is None and \
                            self.is_transfer_protocol_required(error):
                        if payload['Image'].startswith('https://'):
                            payload['TransferProtocolType'] = "HTTPS"
                        elif payload['Image'].startswith('http://'):
                            payload['TransferProtocolType'] = "HTTP"
                        continue

                    # NOTE (iurygregory) we try to handle the case where a
                    # a TransferMethod is also required in the payload.
                    if payload.get('TransferMethod') != "Stream" and \
                            self.is_transfer_method_required(error):
                        payload['TransferMethod'] = "Stream"
                        continue

                    # NOTE (derekh) we try to handle the case where
                    # credentials are required in the payload.
                    # Seen on nvidia dgx (Bug 2071945)
                    if payload.get('UserName') is None and \
                            self.is_credentials_required(error):
                        payload['UserName'] = "none"
                        payload['Password'] = "none"  # noqa:S105
                        continue
                    raise
        self.invalidate()

    def eject_media(self):
        """Detach remote media from virtual media

        After ejecting media inserted will be False and image_name will be
        empty.
        """
        try:
            target_uri, use_patch = self._get_eject_media_uri()
            if use_patch:
                payload = {
                    "Image": None,
                    "Inserted": False
                }
                headers = None
                etag = self._get_etag()
                if etag is not None:
                    headers = {"If-Match": etag}
                self._conn.patch(target_uri, data=payload, headers=headers)
            else:
                self._conn.post(target_uri)
        except exceptions.HTTPError as response:
            # Some vendors like HPE iLO has this kind of implementation.
            # It needs to pass an empty dict.
            if response.status_code in (
                    http_client.UNSUPPORTED_MEDIA_TYPE,
                    http_client.BAD_REQUEST):
                try:
                    self._conn.post(target_uri, data={})
                except exceptions.HTTPError as vmd_exc:
                    if (vmd_exc.status_code >= 500
                        or vmd_exc.status_code in {http_client.BAD_REQUEST,
                                                   http_client.CONFLICT}):
                        LOG.warning('Received exception "%(exception)s" '
                                    'while attempting to eject virtual media:'
                                    ' %(id)s. Refreshing the resource.',
                                    {'exception': vmd_exc,
                                     'id': self.identity})
                        self.refresh(force=True)
                        if not self.inserted:
                            LOG.debug("No image inserted on Virtual Media %s",
                                      self.identity)
                            pass
                        else:
                            raise

        self.invalidate()

    def set_verify_certificate(self, verify_certificate):
        """Enable or disable certificate validation."""
        if not isinstance(verify_certificate, bool):
            raise exceptions.InvalidParameterValueError(
                parameter='verify_certificate', value=verify_certificate,
                valid_values='boolean (True, False)')

        etag = self._get_etag()
        self._conn.patch(self.path,
                         data={'VerifyCertificate': verify_certificate},
                         etag=etag)
        self.invalidate()

    @property
    @utils.cache_it
    def certificates(self):
        """Get the collection of certificates for this device."""
        if not self._certificates_path:
            raise exceptions.MissingAttributeError(
                attribute='Certificates/@odata.id',
                resource=self._path)

        return certificate.CertificateCollection(
            self._conn, self._certificates_path,
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)


class VirtualMediaCollection(base.ResourceCollectionBase):
    """A collection of virtual media attached to a Manager"""

    @property
    def _resource_type(self):
        return VirtualMedia



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock


import sushy
from sushy.resources import constants as res_cons
from sushy.resources.system.storage import controller
from sushy.resources.system.storage import drive
from sushy.resources.system.storage import storage
from sushy.resources.system.storage import volume
from sushy.tests.unit import base


STORAGE_DRIVE_FILE_NAMES = [
    'sushy/tests/unit/json_samples/drive.json',
    'sushy/tests/unit/json_samples/drive2.json',
    'sushy/tests/unit/json_samples/drive3.json'
]

STORAGE_VOLUME_FILE_NAMES = [
    'sushy/tests/unit/json_samples/volume_collection.json',
    'sushy/tests/unit/json_samples/volume.json',
    'sushy/tests/unit/json_samples/volume2.json',
    'sushy/tests/unit/json_samples/volume3.json'
]


class StorageTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            self.json_doc = json.load(f)
        with open('sushy/tests/unit/json_samples/'
                  'storage_controller_collection.json') as f:
            self.json_doc_ctrl_col = json.load(f)
        with open('sushy/tests/unit/json_samples/'
                  'storage_controller.json') as f:
            self.json_doc_ctrl = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.storage = storage.Storage(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.storage._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.storage.redfish_version)
        self.assertEqual('1', self.storage.identity)
        self.assertEqual('Local Storage Controller', self.storage.name)
        self.assertEqual(res_cons.Health.OK, self.storage.status.health)
        self.assertEqual(res_cons.Health.OK, self.storage.status.health_rollup)
        self.assertEqual(res_cons.State.ENABLED, self.storage.status.state)
        self.assertEqual(
            ('/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/35D38F11ACEF7BD3',  # noqa
             '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3F5A8C54207B7233',  # noqa
             '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/32ADF365C6C1B7BD',  # noqa
             '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3D58ECBC375FD9F2',  # noqa
             ), self.storage.drives_identities)

    def test_get_drive(self):
        # | WHEN |
        actual_drive = self.storage.get_drive(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/'
            '35D38F11ACEF7BD3')
        # | THEN |
        self.assertIsInstance(actual_drive, drive.Drive)
        self.assertTrue(self.conn.get.return_value.json.called)

    @mock.patch.object(drive, 'Drive', autospec=True)
    def test_drives(self, Drive_mock):
        # | WHEN |
        all_drives = self.storage.drives
        # | THEN |
        calls = [
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/35D38F11ACEF7BD3',  # noqa
                      redfish_version=self.storage.redfish_version,
                      registries=None, root=None),
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3F5A8C54207B7233',  # noqa
                      redfish_version=self.storage.redfish_version,
                      registries=None, root=None),
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/32ADF365C6C1B7BD',  # noqa
                      redfish_version=self.storage.redfish_version,
                      registries=None, root=None),
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3D58ECBC375FD9F2',  # noqa
                      redfish_version=self.storage.redfish_version,
                      registries=None, root=None)
        ]
        Drive_mock.assert_has_calls(calls)
        self.assertIsInstance(all_drives, list)
        self.assertEqual(4, len(all_drives))
        self.assertIsInstance(all_drives[0], drive.Drive.__class__)

        # returning cached value
        Drive_mock.reset_mock()
        # | WHEN |
        all_drives = self.storage.drives
        # | THEN |
        self.assertFalse(Drive_mock.called)
        self.assertIsInstance(all_drives, list)
        self.assertEqual(4, len(all_drives))
        self.assertIsInstance(all_drives[0], drive.Drive.__class__)

    def test_storage_controllers(self):
        controllers = self.storage.storage_controllers
        self.assertIsInstance(controllers, list)
        self.assertEqual(1, len(controllers))
        controller = controllers[0]
        self.assertEqual('0', controller.member_id)
        self.assertEqual(res_cons.Health.OK, controller.status.health)
        self.assertEqual(res_cons.State.ENABLED, controller.status.state)
        identifiers = controller.identifiers
        self.assertIsInstance(identifiers, list)
        self.assertEqual(1, len(identifiers))
        identifier = identifiers[0]
        self.assertEqual(sushy.DurableNameFormat.NAA,
                         identifier.durable_name_format)
        self.assertEqual('345C59DBD970859C', identifier.durable_name)
        self.assertEqual(12, controller.speed_gbps)
        self.assertEqual([sushy.Protocol.PCIe],
                         controller.controller_protocols)
        self.assertEqual([sushy.Protocol.SAS, sushy.Protocol.SATA],
                         controller.device_protocols)
        self.assertEqual([sushy.RAIDType.RAID0, sushy.RAIDType.RAID1],
                         controller.raid_types)

    def test_controllers(self):
        self.conn.get.return_value.json.side_effect = [
            self.json_doc_ctrl_col,
            self.json_doc_ctrl]
        controllers = self.storage.controllers
        self.assertIsInstance(controllers, controller.ControllerCollection)
        self.assertEqual(1, len(controllers.get_members()))
        controller1 = controllers.get_members()[0]
        self.assertEqual('1', controller1.identity)

    def test_drives_after_refresh(self):
        self.storage.refresh()
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        all_drives = self.storage.drives
        self.assertIsInstance(all_drives, list)
        self.assertEqual(4, len(all_drives))
        for drv in all_drives:
            self.assertIsInstance(drv, drive.Drive)

    def test_drives_max_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.storage.drives_max_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(899527000000, self.storage.drives_max_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_drives_max_size_bytes_after_refresh(self):
        self.storage.refresh()
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.storage.drives_max_size_bytes)

    def test_volumes(self):
        # | GIVEN |
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # | WHEN |
        actual_volumes = self.storage.volumes
        # | THEN |
        self.assertIsInstance(actual_volumes,
                              volume.VolumeCollection)
        self.conn.get.return_value.json.assert_called_once_with()

    def test_volumes_cached(self):
        # | GIVEN |
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # invoke it once
        actual_volumes = self.storage.volumes
        self.conn.get.return_value.json.reset_mock()
        # | WHEN & THEN |
        # tests for same object on invoking subsequently
        self.assertIs(actual_volumes,
                      self.storage.volumes)
        self.conn.get.return_value.json.assert_not_called()

    def test_volumes_on_refresh(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # | WHEN & THEN |
        vols = self.storage.volumes
        self.assertIsInstance(vols, volume.VolumeCollection)

        # On refreshing the system instance...
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.storage.invalidate()
        self.storage.refresh(force=False)

        # | WHEN & THEN |
        self.assertTrue(vols._is_stale)

        # | GIVEN |
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # | WHEN & THEN |
        self.assertIsInstance(self.storage.volumes,
                              volume.VolumeCollection)


class StorageCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'storage_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.stor_col = storage.StorageCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.stor_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Systems/437XR1138R2/Storage/1',),
            self.stor_col.members_identities)

    @mock.patch.object(storage, 'Storage', autospec=True)
    def test_get_member(self, Storage_mock):
        self.stor_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/Storage/1')
        Storage_mock.assert_called_once_with(
            self.stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1',
            redfish_version=self.stor_col.redfish_version, registries=None,
            root=self.stor_col.root)

    @mock.patch.object(storage, 'Storage', autospec=True)
    def test_get_members(self, Storage_mock):
        members = self.stor_col.get_members()
        Storage_mock.assert_called_once_with(
            self.stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1',
            redfish_version=self.stor_col.redfish_version, registries=None,
            root=self.stor_col.root)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))

    def test_drives_sizes_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual([899527000000, 899527000000, 899527000000,
                          899527000000], self.stor_col.drives_sizes_bytes)

    def test_max_drive_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.stor_col.max_drive_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(899527000000, self.stor_col.max_drive_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_drive_size_bytes_after_refresh(self):
        self.stor_col.refresh(force=False)
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.stor_col.max_drive_size_bytes)

    def test_volumes_sizes_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_VOLUME_FILE_NAMES:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual([107374182400, 899527000000, 1073741824000],
                         self.stor_col.volumes_sizes_bytes)

    def test_max_volume_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_VOLUME_FILE_NAMES:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(1073741824000, self.stor_col.max_volume_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(1073741824000, self.stor_col.max_volume_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_volume_size_bytes_after_refresh(self):
        self.stor_col.refresh(force=False)
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_VOLUME_FILE_NAMES:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(1073741824000, self.stor_col.max_volume_size_bytes)

    @mock.patch.object(storage, 'Storage', autospec=True)
    def test_get_members_with_multiple_storages(self, Storage_mock):
        # Test collection with multiple storage members
        with open('sushy/tests/unit/json_samples/'
                  'storage_collection_multiple.json') as f:
            multiple_storages_json = json.loads(f.read())

        self.conn.get.return_value.json.return_value = multiple_storages_json

        # Create a new collection instance to get fresh data
        stor_col_multi = storage.StorageCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage',
            redfish_version='1.0.2')

        members = stor_col_multi.get_members()

        # Should call Storage constructor for each member
        expected_calls = [
            mock.call(self.conn, '/redfish/v1/Systems/437XR1138R2/Storage/1',
                      redfish_version='1.0.2', registries=None, root=None),
            mock.call(self.conn, '/redfish/v1/Systems/437XR1138R2/Storage/2',
                      redfish_version='1.0.2', registries=None, root=None)
        ]
        Storage_mock.assert_has_calls(expected_calls)
        self.assertEqual(2, len(members))

    def test_get_members_expanded(self):
        # Test with expanded JSON data containing full member objects
        # This simulates a response from ?$expand=.($levels=1)
        with open('sushy/tests/unit/json_samples/'
                  'storage_collection_expanded.json') as f:
            expanded_json = json.loads(f.read())

        # Create collection with the URL that would have ?$expand appended
        expanded_collection = storage.StorageCollection(
            self.conn,
            "/redfish/v1/Systems/1/Storage?$expand=.($levels=1)",
            redfish_version="1.0.2",
        )
        expanded_collection._json = expanded_json

        # Mock the Storage objects that will be created
        mock_storage1 = mock.Mock(spec=storage.Storage)
        mock_storage1.identity = "RAID-1"
        mock_storage1.name = "RAID Controller"

        mock_storage2 = mock.Mock(spec=storage.Storage)
        mock_storage2.identity = "NVMe-1"
        mock_storage2.name = "NVMe Storage Controller"

        with mock.patch.object(
            storage, "Storage", autospec=True
        ) as mock_storage_cls:
            # Make the constructor return our mocked instances
            mock_storage_cls.side_effect = [
                mock_storage1,
                mock_storage2,
            ]

            members = expanded_collection.get_members()

            # Verify it created Storage objects from expanded data
            self.assertEqual(2, len(members))
            self.assertEqual([mock_storage1, mock_storage2], members)

            # Check that Storage was initialized with expanded JSON data
            self.assertEqual(2, mock_storage_cls.call_count)

            # Verify first member initialization with expanded data
            first_call = mock_storage_cls.call_args_list[0]
            self.assertEqual(expanded_collection._conn, first_call[0][0])
            self.assertEqual(
                "/redfish/v1/Systems/1/Storage/RAID-1", first_call[0][1]
            )
            # Check that expanded data was passed
            self.assertEqual(
                expanded_json["Members"][0], first_call[1]["json_doc"]
            )

            # Verify second member initialization with expanded data
            second_call = mock_storage_cls.call_args_list[1]
            self.assertEqual(expanded_collection._conn, second_call[0][0])
            self.assertEqual(
                "/redfish/v1/Systems/1/Storage/NVMe-1", second_call[0][1]
            )
            # Check that expanded data was passed
            self.assertEqual(
                expanded_json["Members"][1], second_call[1]["json_doc"]
            )

    def test_get_members_unexpanded(self):
        # Test with unexpanded JSON data containing only references
        # This simulates a normal response without ?$expand parameter
        with open('sushy/tests/unit/json_samples/'
                  'storage_collection.json') as f:
            unexpanded_json = json.loads(f.read())

        # Create collection without expand in URL
        unexpanded_collection = storage.StorageCollection(
            self.conn,
            "/redfish/v1/Systems/1/Storage",
            redfish_version="1.0.2",
        )
        unexpanded_collection._json = unexpanded_json

        # Mock the parent get_members method which will fetch each member
        # individually
        mock_storage1 = mock.Mock(spec=storage.Storage)
        mock_storage2 = mock.Mock(spec=storage.Storage)
        with mock.patch.object(
            storage.base.ResourceCollectionBase,
            "get_members",
            autospec=True,
            return_value=[mock_storage1, mock_storage2],
        ) as mock_super:
            members = unexpanded_collection.get_members()

            # Should fall back to parent implementation for individual fetches
            mock_super.assert_called_once_with(unexpanded_collection)
            self.assertEqual([mock_storage1, mock_storage2], members)

    def test_get_members_no_json(self):
        # Test when collection has no _json attribute
        collection_no_json = storage.StorageCollection(
            self.conn,
            "/redfish/v1/Systems/437XR1138R2/Storage",
            redfish_version="1.0.2",
        )

        # Mock the parent get_members method
        with mock.patch.object(
            storage.base.ResourceCollectionBase,
            "get_members",
            autospec=True,
            return_value=["mock_member"],
        ) as mock_super:
            members = collection_no_json.get_members()

            # Should fall back to parent implementation
            mock_super.assert_called_once()
            self.assertEqual(["mock_member"], members)
